# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli Workspace commands."""

import math
from typing import Any, override
from unittest import mock

import yaml

from debusine.client.commands.base import WorkspaceCommand
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    WorkspaceInheritanceChain,
    WorkspaceInheritanceChainElement,
)
from debusine.utils.input import YamlEditor


class CliWorkspaceInheritanceTests(BaseCliTests):
    """Tests for Cli workspace-inheritance."""

    @override
    def setUp(self) -> None:
        super().setUp()
        build_debusine = self.patch_build_debusine_object()
        self.debusine = build_debusine.return_value

    def cli(
        self,
        args: list[str],
        *,
        current: WorkspaceInheritanceChain,
        expected_stderr: str = "",
    ) -> WorkspaceInheritanceChain | None:
        """
        Run the command with the given arguments.

        :returns: the inheritance chain submitted to the server, or None if
            nothing was submitted
        """
        with (
            mock.patch.object(
                self.debusine,
                "get_workspace_inheritance",
                return_value=current or WorkspaceInheritanceChain(),
            ),
            mock.patch.object(
                self.debusine,
                "set_workspace_inheritance",
                side_effect=lambda _self, chain: chain,
            ) as set_inheritance,
        ):
            cli = self.create_cli(
                ["workspace", "inheritance", "workspace", *args]
            )
            stderr, stdout = self.capture_output(cli.execute)
            self.assertEqual(stderr, expected_stderr)
            if set_inheritance.called:
                submitted = set_inheritance.call_args.kwargs["chain"]
                assert isinstance(submitted, WorkspaceInheritanceChain)
                self.assertEqual(
                    stdout,
                    yaml.safe_dump(
                        submitted.model_dump(),
                        sort_keys=False,
                        width=math.inf,
                    ),
                )
                return submitted
            else:
                self.assertEqual(
                    stdout,
                    yaml.safe_dump(
                        current.model_dump(),
                        sort_keys=False,
                        width=math.inf,
                    ),
                )
                return None

    def chain(
        self,
        *args: WorkspaceInheritanceChainElement,
    ) -> WorkspaceInheritanceChain:
        """Shortcut to build an inheritance chain."""
        return WorkspaceInheritanceChain(chain=list(args))

    def test_get(self) -> None:
        El = WorkspaceInheritanceChainElement
        self.assertIsNone(self.cli([], current=self.chain(El(id=1), El(id=2))))

    def test_get_invalid_workspace(self) -> None:
        with (
            mock.patch.object(
                self.debusine,
                "get_workspace_inheritance",
                side_effect=DebusineError(title="expected error"),
            ),
        ):
            cli = self.create_cli(["workspace", "inheritance", "workspace"])
            stderr, stdout = self.capture_output(
                cli.execute, assert_system_exit_code=3
            )
            self.assertEqual(
                stderr, "result: failure\nerror:\n  title: expected error\n"
            )
            self.assertEqual(stdout, "")

    def test_change(self) -> None:
        El = WorkspaceInheritanceChainElement
        el1, el2, el3 = El(id=1), El(id=2), El(id=3)
        for current, args, expected in (
            (
                self.chain(el2),
                ["--append", "1", "3"],
                self.chain(el2, el1, el3),
            ),
            (self.chain(), ["--append", "1", "3"], self.chain(el1, el3)),
            (
                self.chain(el2),
                ["--prepend", "1", "3"],
                self.chain(el1, el3, el2),
            ),
            (self.chain(), ["--prepend", "1", "3"], self.chain(el1, el3)),
            (
                self.chain(el1, el2, el3),
                ["--remove", "1"],
                self.chain(el2, el3),
            ),
            (
                self.chain(el1, el2, el3),
                ["--set", "3", "1"],
                self.chain(el3, el1),
            ),
            (
                self.chain(el1),
                ["--append", "2", "--prepend", "3"],
                self.chain(el3, el1, el2),
            ),
            (
                self.chain(el1),
                ["--append", "2", "--set", "3"],
                self.chain(el3),
            ),
            (
                self.chain(el2),
                ["--prepend", "1", "3"],
                self.chain(el1, el3, el2),
            ),
        ):
            with self.subTest(current=repr(current), args=args):
                self.assertEqual(
                    self.cli(args, current=current),
                    expected,
                )

    def test_change_invalid_chain(self) -> None:
        El = WorkspaceInheritanceChainElement
        with (
            mock.patch.object(
                self.debusine,
                "get_workspace_inheritance",
                return_value=WorkspaceInheritanceChain(),
            ),
            mock.patch.object(
                self.debusine,
                "set_workspace_inheritance",
                side_effect=DebusineError(title="expected error"),
            ) as set_inheritance,
        ):
            cli = self.create_cli(
                ["workspace", "inheritance", "workspace", "--set", "1", "1"]
            )
            stderr, stdout = self.capture_output(
                cli.execute, assert_system_exit_code=3
            )
            self.assertEqual(stdout, "")
            self.assertEqual(
                stderr,
                "Error setting inheritance:\n\n"
                "result: failure\n"
                "error:\n"
                "  title: expected error\n",
            )
            set_inheritance.assert_called_with(
                "workspace", chain=self.chain(El(id=1), El(id=1))
            )

    def test_edit(self) -> None:
        def mock_edit(self_: YamlEditor[list[Any]]) -> bool:
            self_.value = self.chain(el3, el2, el1).model_dump()["chain"]
            self_.cleanup()
            return True

        El = WorkspaceInheritanceChainElement
        el1, el2, el3 = El(id=1), El(id=2), El(id=3)
        with mock.patch(
            "debusine.utils.input.YamlEditor.edit",
            side_effect=mock_edit,
            autospec=True,
        ):
            self.assertEqual(
                self.cli(["--edit"], current=self.chain(el1, el2, el3)),
                self.chain(el3, el2, el1),
            )

    def test_edit_old_new_same(self) -> None:
        def mock_edit(self_: YamlEditor[list[Any]]) -> bool:
            self_.value = self.chain(el1, el2, el3).model_dump()["chain"]
            self_.cleanup()
            return True

        El = WorkspaceInheritanceChainElement
        el1, el2, el3 = El(id=1), El(id=2), El(id=3)
        with mock.patch(
            "debusine.utils.input.YamlEditor.edit",
            side_effect=mock_edit,
            autospec=True,
        ):
            # self.cli returns None if Debusine.set_inheritance is not called
            # (it's not called because the user didn't try changing the
            # inheritance chain)
            self.assertIsNone(
                self.cli(["--edit"], current=self.chain(el1, el2, el3)),
            )

    def test_edit_failed_editor_returns_error(self) -> None:
        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            self_.cleanup()
            # Editor failed (e.g. user never wrote the expected YAML data
            # structure)
            return False

        El = WorkspaceInheritanceChainElement
        el1, el2, el3 = El(id=1), El(id=2), El(id=3)
        with mock.patch(
            "debusine.utils.input.YamlEditor.edit",
            side_effect=mock_edit,
            autospec=True,
        ):
            self.assertIsNone(
                self.cli(["--edit"], current=self.chain(el1, el2, el3))
            )

    def test_edit_failed_apply_inheritance_failed_one_retry(self) -> None:
        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            self_.value = self.chain(
                El(workspace="does-not-exist")
            ).model_dump()["chain"]
            self_.cleanup()
            return True

        El = WorkspaceInheritanceChainElement
        with (
            mock.patch(
                "debusine.utils.input.YamlEditor.edit",
                side_effect=mock_edit,
                autospec=True,
            ),
            mock.patch(
                "debusine.client.commands."
                "workspaces.Inheritance._apply_inheritance",
                side_effect=DebusineError(
                    title="Workspace name does not exist"
                ),
                autospec=True,
            ),
            mock.patch(
                "debusine.utils.input.YamlEditor.confirm",
                # Re-try once, then give up
                side_effect=[True, False],
                autospec=True,
            ),
        ):
            expected_stderr = (
                "Error setting workspace inheritance:\n"
                "\n"
                "result: failure\n"
                "error:\n"
                "  title: Workspace name does not exist\n"
            )
            self.assertIsNone(
                self.cli(
                    ["--edit"],
                    current=self.chain(El(workspace="devel")),
                    expected_stderr=expected_stderr * 2,
                )
            )

    def test_edit_api_call_failed(self) -> None:
        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            self_.value = self.chain(el3, el2, el1).model_dump()["chain"]
            self_.cleanup()
            return True

        El = WorkspaceInheritanceChainElement
        el1, el2, el3 = El(id=1), El(id=2), El(id=3)
        with (
            mock.patch(
                "debusine.utils.input.YamlEditor.edit",
                side_effect=mock_edit,
                autospec=True,
            ),
            mock.patch(
                "debusine.client.commands."
                "workspaces.Inheritance._apply_inheritance",
                side_effect=[
                    DebusineError(title="expected error"),
                    WorkspaceInheritanceChain(),
                ],
                autospec=True,
            ),
            mock.patch(
                "debusine.utils.input.YamlEditor.confirm",
                side_effect=[True, False],
                autospec=True,
            ),
        ):
            expected_stderr = (
                "Error setting workspace inheritance:\n"
                "\n"
                "result: failure\n"
                "error:\n"
                "  title: expected error\n"
            )
            self.assertIsNone(
                self.cli(
                    ["--edit"],
                    current=self.chain(),
                    expected_stderr=expected_stderr,
                )
            )

    def test_missing_legacy_workspace_uses_standard_one(self) -> None:
        command = self.create_command(
            ["workspace", "inheritance", "--workspace=name"]
        )
        assert isinstance(command, WorkspaceCommand)
        self.assertEqual(command.workspace, "name")

    def test_missing_legacy_workspace_uses_standard_one_default(self) -> None:
        command = self.create_command(["workspace", "inheritance"])
        assert isinstance(command, WorkspaceCommand)
        self.assertEqual(command.workspace, "developers")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used by debusine signing tasks."""

from typing import Annotated, override

import pydantic

from debusine.artifacts.models import WorkRequestResults
from debusine.assets import KeyPurpose
from debusine.signing.models import SigningMode
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseTaskData,
    LookupMultiple,
    LookupSingle,
)


class SigningNoopData(BaseTaskData):
    """In-memory task data for the SigningNoop task."""

    result: WorkRequestResults = WorkRequestResults.FAILURE


class GenerateKeyData(BaseTaskData):
    """In-memory task data for the GenerateKey task."""

    purpose: KeyPurpose
    description: str


class SignData(BaseTaskData):
    """In-memory task data for the Sign task."""

    purpose: KeyPurpose
    unsigned: LookupMultiple
    key: str

    @pydantic.field_validator("purpose", mode="after")
    @classmethod
    def validate_purpose(cls, purpose: KeyPurpose) -> KeyPurpose:
        """Don't allow OpenPGP repository signing keys."""
        if purpose == KeyPurpose.OPENPGP_REPOSITORY:
            raise ValueError(
                f"Key purpose may not be '{KeyPurpose.OPENPGP_REPOSITORY}'"
                f" here (handled by SignRepositoryIndex task instead)"
            )
        return purpose


class SignDynamicData(BaseDynamicTaskData):
    """Dynamic data for the Sign task."""

    unsigned_ids: list[int]
    unsigned_binary_package_names: list[str | None] = []

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return self.unsigned_ids


class DebsignData(BaseTaskData):
    """In-memory task data for the Debsign task."""

    unsigned: LookupSingle
    key: str


class DebsignDynamicData(BaseDynamicTaskData):
    """Dynamic data for the Debsign task."""

    unsigned_id: int

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [self.unsigned_id]


class SignRepositoryIndexData(BaseTaskData):
    """In-memory task data for the SignRepositoryIndex task."""

    suite_collection: LookupSingle
    unsigned: LookupSingle
    mode: SigningMode
    signed_name: Annotated[str, pydantic.Field(pattern=r"^[^/]+$")]

    @pydantic.field_validator("mode", mode="after")
    @classmethod
    def validate_mode(cls, mode: SigningMode) -> SigningMode:
        """Only allow detached or clear signatures."""
        if mode not in {SigningMode.DETACHED, SigningMode.CLEAR}:
            raise ValueError(
                f"Signing mode must be 'detached' or 'clear'; got '{mode}'"
            )
        return mode


class SignRepositoryIndexDynamicData(BaseDynamicTaskData):
    """Dynamic data for the SignRepositoryIndex task."""

    signing_keys: list[str]
    unsigned_id: int

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [self.unsigned_id]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the autopkgtest workflow."""

import datetime as dt
from collections.abc import Iterable, Sequence
from typing import ClassVar, TypeAlias, override
from unittest import mock

from django.test import override_settings
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianAutopkgtest,
    DebianAutopkgtestResult,
    DebianAutopkgtestResultStatus,
    DebianAutopkgtestSource,
    DebusinePromise,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, Collection, CollectionItem, WorkRequest
from debusine.db.models.work_requests import SkipWorkRequest
from debusine.server.workflows import (
    AutopkgtestWorkflow,
    SbuildWorkflow,
    WorkflowValidationError,
    workflow_utils,
)
from debusine.server.workflows.models import (
    AutopkgtestWorkflowData,
    AutopkgtestWorkflowDynamicData,
    BaseWorkflowData,
    RegressionTrackingUpdateQAResults,
    SbuildWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.server.workflows.tests.test_regression_tracking import (
    RegressionTrackingWorkflowTestMixin,
)
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    ActionUpdateCollectionWithArtifacts,
    AutopkgtestNeedsInternet,
    BackendType,
    BaseDynamicTaskData,
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    OutputData,
    RegressionAnalysisStatus,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class AutopkgtestWorkflowTests(
    RegressionTrackingWorkflowTestMixin, WorkflowTestCase[AutopkgtestWorkflow]
):
    """Unit tests for :py:class:`AutopkgtestWorkflow`."""

    SAMPLE_TASK_DATA = {
        "vendor": "debian",
        "codename": "sid",
    }

    source_artifact: ClassVar[Artifact]
    binary_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        cls.binary_artifact = (
            cls.playground.create_minimal_binary_package_artifact()
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["source_artifact"] = self.source_artifact.pk
        self.task_data["binary_artifacts"] = [self.binary_artifact.pk]

    def create_binary_upload(
        self, architecture: str, filenames: list[str]
    ) -> Artifact:
        """Create a minimal `debian:upload` artifact with binaries."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": architecture,
                    "Files": [{"name": filename} for filename in filenames],
                },
            },
        )
        return artifact

    def add_uploads(
        self, work_request: WorkRequest, architectures: Iterable[str]
    ) -> None:
        """Add multiple uploads to a workflow's internal collection."""
        internal_collection = work_request.internal_collection
        assert internal_collection is not None
        for architecture in architectures:
            upload = self.create_binary_upload(
                architecture, [f"hello_1.0-1_{architecture}.deb"]
            )
            internal_collection.manager.add_artifact(
                upload,
                user=work_request.created_by,
                workflow=work_request,
                name=f"build-{architecture}",
                variables={"architecture": architecture},
            )

    def add_qa_result(
        self,
        qa_results: Collection,
        package: str,
        version: str,
        architecture: str,
        *,
        timestamp: dt.datetime | None = None,
    ) -> CollectionItem:
        """Add an autopkgtest result to a ``debian:qa-results`` collection."""
        work_request = self.playground.create_worker_task(
            task_name="autopkgtest",
            result=WorkRequest.Results.SUCCESS,
            validate=False,
            use_data=False,
        )
        return qa_results.manager.add_artifact(
            self.playground.create_artifact(
                category=ArtifactCategory.AUTOPKGTEST
            )[0],
            user=self.playground.get_default_user(),
            variables={
                "package": package,
                "version": version,
                "architecture": architecture,
                "timestamp": int((timestamp or timezone.now()).timestamp()),
                "work_request_id": work_request.id,
            },
        )

    def orchestrate(
        self,
        task_data: AutopkgtestWorkflowData,
        architectures: Sequence[str],
        parent: WorkRequest | None = None,
        pipeline_task_name: str = "examplepipeline",
    ) -> WorkRequest:
        """Create and orchestrate an AutopkgtestWorkflow."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Pipeline workflow that runs sbuild and autopkgtest."""

            TASK_NAME = pipeline_task_name

            @override
            def populate(self_) -> None:
                sbuild = self_.work_request.create_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(
                            source_artifact=task_data.source_artifact
                        ),
                        target_distribution="debian:sid",
                        architectures=["all", *architectures],
                    ),
                )
                sbuild_workflow = sbuild.get_task()
                assert isinstance(sbuild_workflow, SbuildWorkflow)
                source_artifact = (
                    workflow_utils.locate_debian_source_package_lookup(
                        sbuild_workflow,
                        "input.source_artifact",
                        task_data.source_artifact,
                    )
                )
                self.playground.advance_work_request(sbuild, mark_running=True)
                for architecture in ("all", *architectures):
                    child = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(source_artifact=source_artifact),
                            build_architecture=architecture,
                            environment="debian/match:codename=sid",
                        ),
                    )
                    self_.provides_artifact(
                        child,
                        ArtifactCategory.UPLOAD,
                        f"build-{architecture}",
                        data={"architecture": architecture},
                    )
                sbuild.unblock_workflow_children()

                autopkgtest = self_.work_request.create_child_workflow(
                    task_name="autopkgtest",
                    task_data=task_data,
                )
                self.playground.advance_work_request(
                    autopkgtest, mark_pending=True
                )
                self_.orchestrate_child(autopkgtest)

        root = self.playground.create_workflow(
            task_name=pipeline_task_name, parent=parent
        )
        self.assertTrue(
            self.playground.orchestrate_workflow(root, use_data=False)
        )

        return root

    def test_create_orchestrator(self) -> None:
        """An AutopkgtestWorkflow can be instantiated."""
        source_artifact = self.source_artifact.pk
        binary_artifacts = ("internal@collections/name:build-arm64",)
        vendor = "debian"
        codename = "trixie"
        self.task_data.update(
            {
                "source_artifact": source_artifact,
                "binary_artifacts": binary_artifacts,
                "vendor": vendor,
                "codename": codename,
            }
        )
        self.validate = False

        w = self.task
        self.assertEqual(w.data.source_artifact, source_artifact)
        self.assertEqual(
            w.data.binary_artifacts, LookupMultiple(binary_artifacts)
        )
        self.assertEqual(w.data.vendor, vendor)
        self.assertEqual(w.data.codename, codename)
        self.assertEqual(w.data.backend, BackendType.UNSHARE)

    def test_create_orchestrator_explicit_backend(self) -> None:
        """An AutopkgtestWorkflow can be instantiated with a backend."""
        self.task_data["backend"] = BackendType.INCUS_LXC
        self.validate = False
        self.assertEqual(self.task.data.backend, BackendType.INCUS_LXC)

    def test_architectures(self) -> None:
        """Workflow uses architectures from binary_artifacts."""
        architectures = ("amd64", "arm64", "i386")
        self.task_data["binary_artifacts"] = [
            f"internal@collections/name:build-{architecture}"
            for architecture in architectures
        ]
        self.validate = False
        self.add_uploads(self.work_request, architectures)
        self.assertCountEqual(self.task.architectures, architectures)

    def test_architectures_arch_indep_and_arch_dep(self) -> None:
        """Workflow handles arch-indep plus arch-dep binary_artifacts."""
        architectures = ("all", "arm64", "armhf")
        self.task_data["binary_artifacts"] = [
            f"internal@collections/name:build-{architecture}"
            for architecture in architectures
        ]
        self.validate = False
        self.add_uploads(self.work_request, architectures)
        self.assertEqual(self.task.architectures, {"arm64", "armhf"})

    def test_architectures_arch_indep_only(self) -> None:
        """Workflow handles only having arch-indep binary_artifacts."""
        self.task_data["binary_artifacts"] = [
            "internal@collections/name:build-all"
        ]
        self.validate = False
        self.add_uploads(self.work_request, ["all"])
        self.assertEqual(self.task.architectures, {"amd64"})

    def test_architectures_arch_indep_only_arch_all_build_architecture(
        self,
    ) -> None:
        """Workflow honours `arch_all_build_architecture`."""
        self.task_data["binary_artifacts"] = [
            "internal@collections/name:build-all"
        ]
        self.task_data["arch_all_build_architecture"] = "s390x"
        self.validate = False
        self.add_uploads(self.work_request, ["all"])
        self.assertEqual(self.task.architectures, {"s390x"})

    def test_architectures_intersect_task_data(self) -> None:
        """Setting architectures in task data constrains the set."""
        architectures = ("amd64", "arm64", "i386")
        self.task_data["binary_artifacts"] = [
            f"internal@collections/name:build-{architecture}"
            for architecture in architectures
        ]
        self.task_data["architectures"] = ["amd64", "i386"]
        self.validate = False
        self.add_uploads(self.work_request, architectures)
        self.assertEqual(self.task.architectures, {"amd64", "i386"})

    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        self.validate = False
        self.add_uploads(self.work_request, ("amd64",))
        self.task.validate_input()

    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        self.task_data["qa_suite"] = "nonexistent@debian:suite"
        self.validate = False
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            self.task.validate_input()

    def test_validate_input_bad_reference_qa_results(self) -> None:
        """validate_input raises errors in looking up reference QA results."""
        self.task_data["reference_qa_results"] = "nonexistent@debian:qa-results"
        self.validate = False
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:qa-results' does not exist or is hidden",
        ):
            self.task.validate_input()

    def test_validate_input_architecture_errors(self) -> None:
        """validate_input raises errors in computing architectures."""
        self.task_data["binary_artifacts"] = [
            "internal@collections/name:build-amd64"
        ]
        self.validate = False
        assert self.work_request.internal_collection is not None
        # Create a promise with no architecture.
        self.playground.create_bare_data_item(
            self.work_request.internal_collection,
            "build-amd64",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=self.work_request.id + 1,
                promise_workflow_id=self.work_request.id,
                promise_category=ArtifactCategory.UPLOAD,
            ),
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "Cannot determine architecture for lookup result",
        ):
            self.task.validate_input()

    def test_has_current_reference_qa_result_no_match(self) -> None:
        """_has_current_reference_qa_result: no matching result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "other", "1.0-1", "amd64")
        self.task_data = {
            "prefix": "reference-qa-result|",
            "source_artifact": self.source_artifact.id,
            "binary_artifacts": LookupMultiple(()),
            "vendor": "debian",
            "codename": "sid",
            "qa_suite": f"sid@{CollectionCategory.SUITE}",
            "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
            "update_qa_results": RegressionTrackingUpdateQAResults.YES,
        }
        self.assertFalse(self.task._has_current_reference_qa_result("amd64"))

    def test_has_current_reference_qa_result_different_version(self) -> None:
        """_has_current_reference_qa_result: result for different version."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )
        self.task_data = {
            "prefix": "reference-qa-result|",
            "source_artifact": source_artifact.id,
            "binary_artifacts": LookupMultiple(()),
            "vendor": "debian",
            "codename": "sid",
            "qa_suite": f"sid@{CollectionCategory.SUITE}",
            "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
            "update_qa_results": RegressionTrackingUpdateQAResults.YES,
        }
        self.assertFalse(self.task._has_current_reference_qa_result("amd64"))

    def test_has_current_reference_qa_result_too_old(self) -> None:
        """_has_current_reference_qa_result: result is too old."""
        now = timezone.now()
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            timestamp=now - dt.timedelta(days=31),
        )
        source_item = sid_suite.manager.add_artifact(
            self.source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        source_item.created_at = now
        source_item.save()

        self.task_data = {
            "prefix": "reference-qa-result|",
            "source_artifact": self.source_artifact.id,
            "binary_artifacts": LookupMultiple(()),
            "vendor": "debian",
            "codename": "sid",
            "qa_suite": f"sid@{CollectionCategory.SUITE}",
            "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
            "update_qa_results": RegressionTrackingUpdateQAResults.YES,
        }
        self.assertFalse(self.task._has_current_reference_qa_result("amd64"))

    def test_has_current_reference_qa_result_current(self) -> None:
        """_has_current_reference_qa_result: current result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")

        self.task_data = {
            "prefix": "reference-qa-result|",
            "source_artifact": self.source_artifact.id,
            "binary_artifacts": LookupMultiple(()),
            "vendor": "debian",
            "codename": "sid",
            "qa_suite": f"sid@{CollectionCategory.SUITE}",
            "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
            "update_qa_results": RegressionTrackingUpdateQAResults.YES,
        }

        self.assertTrue(self.task._has_current_reference_qa_result("amd64"))

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        """The workflow populates child work requests."""
        architectures = ("amd64", "i386")

        root = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                source_artifact=self.source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in ("all", *architectures)
                    ]
                ),
                vendor="debian",
                codename="sid",
                extra_repositories=[
                    ExtraExternalRepository(
                        url="http://example.com/",
                        suite="bookworm",
                        components=["main"],
                    )
                ],
            ),
            architectures=architectures,
        )

        autopkgtest = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest", parent=root
        )
        children = list(
            WorkRequest.objects.filter(parent=autopkgtest).order_by("id")
        )
        self.assertEqual(len(children), len(architectures))
        for child, architecture in zip(children, architectures):
            self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
            self.assertEqual(child.task_type, TaskTypes.WORKER)
            self.assertEqual(child.task_name, "autopkgtest")
            self.assertEqual(
                child.task_data,
                {
                    "input": {
                        "source_artifact": autopkgtest.task_data[
                            "source_artifact"
                        ],
                        "binary_artifacts": sorted(
                            [
                                f"internal@collections/"
                                f"name:build-{architecture}",
                                "internal@collections/name:build-all",
                            ]
                        ),
                        "context_artifacts": [],
                    },
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=sid",
                    "extra_repositories": [
                        {
                            "url": "http://example.com/",
                            "suite": "bookworm",
                            "components": ["main"],
                        }
                    ],
                    "backend": BackendType.UNSHARE,
                    "include_tests": [],
                    "exclude_tests": [],
                    "debug_level": 0,
                    "pin_packages": {},
                    "extra_environment": {},
                    "needs_internet": AutopkgtestNeedsInternet.RUN,
                    "fail_on": {},
                    "timeout": None,
                },
            )
            self.assert_work_request_event_reactions(
                child,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        collection="internal@collections",
                        name_template=f"autopkgtest-{architecture}",
                        artifact_filters={
                            "category": ArtifactCategory.AUTOPKGTEST
                        },
                    )
                ],
            )
            self.assertQuerySetEqual(
                child.dependencies.all(),
                list(
                    WorkRequest.objects.filter(
                        task_type=TaskTypes.WORKER,
                        task_name="sbuild",
                        task_data__build_architecture__in={architecture, "all"},
                    )
                ),
            )
            self.assertEqual(
                child.workflow_data_json,
                {
                    "display_name": f"autopkgtest {architecture}",
                    "step": f"autopkgtest-{architecture}",
                },
            )

        # Population is idempotent.
        self.get_workflow(autopkgtest).populate()
        children = list(WorkRequest.objects.filter(parent=autopkgtest))
        self.assertEqual(len(children), len(architectures))

    @preserve_db_task_registry()
    def test_populate_upload(self) -> None:
        """The workflow accepts debian:upload source artifacts."""
        architectures = ("amd64", "i386")
        upload_artifacts = self.playground.create_upload_artifacts()

        root = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in ("all", *architectures)
                    ]
                ),
                vendor="debian",
                codename="sid",
                extra_repositories=[
                    ExtraExternalRepository(
                        url="http://example.com/",
                        suite="bookworm",
                        components=["main"],
                    )
                ],
            ),
            architectures=architectures,
        )

        autopkgtest = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="autopkgtest",
            parent=root,
        )
        children = list(
            WorkRequest.objects.filter(parent=autopkgtest).order_by("id")
        )
        self.assertEqual(len(children), len(architectures))
        for child, architecture in zip(children, architectures):
            self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
            self.assertEqual(child.task_type, TaskTypes.WORKER)
            self.assertEqual(child.task_name, "autopkgtest")
            self.assertEqual(
                child.task_data,
                {
                    "input": {
                        "source_artifact": (
                            f"{upload_artifacts.source.id}@artifacts"
                        ),
                        "binary_artifacts": sorted(
                            [
                                f"internal@collections/"
                                f"name:build-{architecture}",
                                "internal@collections/name:build-all",
                            ]
                        ),
                        "context_artifacts": [],
                    },
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=sid",
                    "extra_repositories": [
                        {
                            "url": "http://example.com/",
                            "suite": "bookworm",
                            "components": ["main"],
                        }
                    ],
                    "backend": BackendType.UNSHARE,
                    "include_tests": [],
                    "exclude_tests": [],
                    "debug_level": 0,
                    "pin_packages": {},
                    "extra_environment": {},
                    "needs_internet": AutopkgtestNeedsInternet.RUN,
                    "fail_on": {},
                    "timeout": None,
                },
            )
            self.assert_work_request_event_reactions(
                child,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        collection="internal@collections",
                        name_template=f"autopkgtest-{architecture}",
                        artifact_filters={
                            "category": ArtifactCategory.AUTOPKGTEST
                        },
                    )
                ],
            )
            self.assertQuerySetEqual(
                child.dependencies.all(),
                list(
                    WorkRequest.objects.filter(
                        task_type=TaskTypes.WORKER,
                        task_name="sbuild",
                        task_data__build_architecture__in={architecture, "all"},
                    )
                ),
            )
            self.assertEqual(
                child.workflow_data_json,
                {
                    "display_name": f"autopkgtest {architecture}",
                    "step": f"autopkgtest-{architecture}",
                },
            )

        # Population is idempotent.
        self.get_workflow(autopkgtest).populate()
        children = list(WorkRequest.objects.filter(parent=autopkgtest))
        self.assertEqual(len(children), len(architectures))

    def test_populate_mixed_uploads(self) -> None:
        """The workflow accepts mixed debian:upload artifacts in all inputs."""
        artifacts = self.playground.create_upload_artifacts(
            binaries=[
                ("hello", "amd64"),
                ("hello", "i386"),
                ("hello-doc", "all"),
            ]
        )
        context = self.playground.create_upload_artifacts(
            binaries=[
                ("context", "amd64"),
                ("context", "i386"),
                ("context-common", "all"),
            ]
        )
        architectures = ["amd64", "i386"]

        self.task_data.update(
            {
                "source_artifact": artifacts.upload.id,
                "binary_artifacts": [artifacts.upload.id],
                "context_artifacts": [context.upload.id],
                "architectures": architectures,
            }
        )
        self.task.populate()
        children = list(self.work_request.children.order_by("id").all())
        self.assertEqual(len(children), len(architectures))
        for architecture, child in zip(architectures, children):
            with self.subTest(architecture=architecture):
                self.assertEqual(
                    child.task_data["build_architecture"], architecture
                )
                self.assertEqual(
                    child.task_data["input"],
                    {
                        "source_artifact": (f"{artifacts.source.id}@artifacts"),
                        "binary_artifacts": sorted(
                            f"{artifact.id}@artifacts"
                            for artifact in artifacts.binaries
                            if artifact.data["deb_fields"]["Architecture"]
                            in (architecture, "all")
                        ),
                        "context_artifacts": sorted(
                            f"{artifact.id}@artifacts"
                            for artifact in context.binaries
                            if artifact.data["deb_fields"]["Architecture"]
                            in (architecture, "all")
                        ),
                    },
                )

    def test_populate_experimental(self) -> None:
        """The workflow handles overlay distributions."""
        self.playground.create_debian_environment(
            codename="experimental", variant="autopkgtest"
        )
        self.task_data["codename"] = "experimental"
        self.task.populate()
        children = list(self.work_request.children.all())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(len(child.task_data["extra_repositories"]), 1)
        repo = child.task_data["extra_repositories"][0]
        self.assertEqual(repo["suite"], "experimental")

    def test_populate_pin_packages(self) -> None:
        self.task_data["pin_packages"] = {"unstable": ["src:glibc", "hello"]}
        self.task.populate()
        children = list(self.work_request.children.all())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(
            child.task_data["pin_packages"],
            {"unstable": ["src:glibc", "hello"]},
        )

    def test_populate_has_current_reference_qa_result(self) -> None:
        """The workflow does nothing with a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        self.task_data.update(
            {
                "prefix": "reference-qa-result|",
                "binary_artifacts": [binary_artifact.id],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
            }
        )

        self.task.populate()
        self.assertQuerySetEqual(self.work_request.children.all(), [])

    def test_populate_has_current_reference_qa_result_force(self) -> None:
        """The workflow can be forced despite a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        self.task_data.update(
            {
                "prefix": "reference-qa-result|",
                "binary_artifacts": [binary_artifact.id],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.FORCE,
            }
        )

        self.task.populate()

        self.assertTrue(
            self.work_request.children.filter(
                task_type=TaskTypes.WORKER, task_name="autopkgtest"
            ).exists()
        )

    def test_populate_previous_reference_qa_result_too_old(self) -> None:
        """The workflow produces a reference QA result if it is outdated."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        old_qa_result = self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            timestamp=timezone.now() - dt.timedelta(days=31),
        )

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        source_item = sid_suite.manager.add_artifact(
            self.source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        self.task_data.update(
            {
                "prefix": "reference-qa-result|",
                "binary_artifacts": [binary_artifact.id],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "context_artifacts": [],
            }
        )
        self.task.populate()
        [child] = self.work_request.children.all()
        self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "autopkgtest")
        self.assertEqual(
            child.task_data,
            {
                "input": {
                    "source_artifact": self.source_artifact.id,
                    "binary_artifacts": [binary_artifact.id],
                    "context_artifacts": [],
                },
                "build_architecture": "amd64",
                "environment": "debian/match:codename=sid",
                "extra_repositories": None,
                "backend": BackendType.UNSHARE,
                "include_tests": [],
                "exclude_tests": [],
                "debug_level": 0,
                "pin_packages": {},
                "extra_environment": {},
                "needs_internet": AutopkgtestNeedsInternet.RUN,
                "fail_on": {},
                "timeout": None,
            },
        )
        provide_action = ActionUpdateCollectionWithArtifacts(
            collection="internal@collections",
            name_template="reference-qa-result|autopkgtest-amd64",
            artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
        )
        skip_action = ActionSkipIfLookupResultChanged(
            lookup=(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:autopkgtest_hello_amd64"
            ),
            collection_item_id=old_qa_result.id,
            promise_name="reference-qa-result|autopkgtest-amd64",
        )
        qa_result_action = ActionUpdateCollectionWithArtifacts(
            collection=f"sid@{CollectionCategory.QA_RESULTS}",
            variables={
                "package": "hello",
                "version": "1.0-1",
                "architecture": "amd64",
                "timestamp": int(source_item.created_at.timestamp()),
                "work_request_id": child.id,
            },
            artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
        )
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[skip_action],
            on_failure=[qa_result_action],
            on_success=[provide_action, qa_result_action],
        )

        # Completing the work request stores the QA result.
        result, _ = self.playground.create_artifact(
            category=ArtifactCategory.AUTOPKGTEST, work_request=child
        )
        self.playground.advance_work_request(
            child, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:autopkgtest_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).artifact,
            result,
        )

        # Population is idempotent, even if qa_suite changes.
        with (
            mock.patch.object(
                AutopkgtestWorkflow,
                "_has_current_reference_qa_result",
                return_value=False,
            ),
            mock.patch.object(
                AutopkgtestWorkflow,
                "qa_suite_changed",
                mock.PropertyMock(
                    return_value=timezone.now() + dt.timedelta(hours=1)
                ),
            ),
        ):
            self.task.populate()
        self.assertEqual(self.work_request.children.count(), 1)
        child.refresh_from_db()
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[skip_action],
            on_failure=[qa_result_action],
            on_success=[provide_action, qa_result_action],
        )

    def test_populate_reference_qa_result_backs_off(self) -> None:
        """Reference tasks are skipped if another workflow got there first."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.playground.create_debian_environment(
            codename="sid", variant="autopkgtest"
        )

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        self.task_data.update(
            {
                "prefix": "reference-qa-result|",
                "binary_artifacts": LookupMultiple((binary_artifact.id,)),
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
            }
        )
        self.task.populate()

        racing_qa_result = self.add_qa_result(
            sid_qa_results, "hello", "1.0-1", "amd64"
        )
        [child] = self.work_request.children.all()
        self.playground.advance_work_request(child, mark_pending=True)

        with self.assertRaises(SkipWorkRequest):
            child.assign_worker(self.playground.create_worker())

        self.assertEqual(child.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(child.result, WorkRequest.Results.SKIPPED)
        self.assertEqual(
            child.output_data,
            OutputData(
                skip_reason=(
                    f"Result of lookup "
                    f"'sid@{CollectionCategory.QA_RESULTS}/"
                    f"latest:autopkgtest_hello_amd64' changed"
                )
            ),
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:autopkgtest_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).collection_item,
            racing_qa_result,
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_callback_regression_analysis(self) -> None:
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)
        architectures = [
            "amd64",
            "arm64",
            "i386",
            "ppc64el",
            "riscv64",
            "s390x",
        ]
        reference_source_artifact = self.source_artifact
        new_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                architectures=architectures,
                source_artifact=reference_source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                        if arch != "ppc64el"
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            architectures=architectures,
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        new = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                reference_prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                enable_regression_tracking=True,
                architectures=architectures,
                source_artifact=new_source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            architectures=architectures,
            parent=root,
        )

        # Unblock the autopkgtest tasks.
        for sbuild in reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )
        for sbuild in new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )

        DARS: TypeAlias = DebianAutopkgtestResultStatus

        reference_autopkgtest_workflow = reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        reference_source_url = (
            f"http://example.com/{reference_source_artifact.get_absolute_url()}"
        )
        reference_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in reference_autopkgtest_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="autopkgtest"
            )
        }
        reference_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.AUTOPKGTEST,
                data=DebianAutopkgtest(
                    results={
                        name: DebianAutopkgtestResult(status=status)
                        for name, status in statuses.items()
                    },
                    cmdline="unused",
                    source_package=DebianAutopkgtestSource(
                        name="hello", version="1.0-1", url=reference_source_url
                    ),
                    architecture=architecture,
                    distribution="sid",
                ),
                work_request=reference_tasks[architecture],
            )[0]
            for architecture, statuses in (
                ("amd64", {"upstream": DARS.PASS, "old": DARS.PASS}),
                ("arm64", {"upstream": DARS.PASS}),
                ("i386", {"upstream": DARS.FAIL}),
                ("riscv64", {"upstream": DARS.FAIL, "performance": DARS.FAIL}),
                ("s390x", {"upstream": DARS.FAIL}),
            )
        }
        new_autopkgtest_workflow = new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        new_source_url = (
            f"http://example.com/{new_source_artifact.get_absolute_url()}"
        )
        new_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in new_autopkgtest_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="autopkgtest"
            )
        }
        new_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.AUTOPKGTEST,
                data=DebianAutopkgtest(
                    results={
                        name: DebianAutopkgtestResult(status=status)
                        for name, status in statuses.items()
                    },
                    cmdline="unused",
                    source_package=DebianAutopkgtestSource(
                        name="hello", version="1.0-2", url=new_source_url
                    ),
                    architecture=architecture,
                    distribution="sid",
                ),
                work_request=new_tasks[architecture],
            )[0]
            for architecture, statuses in (
                ("amd64", {"upstream": DARS.PASS}),
                ("arm64", {"upstream": DARS.FAIL}),
                ("i386", {"upstream": DARS.PASS}),
                ("ppc64el", {"new": DARS.PASS}),
                ("riscv64", {"upstream": DARS.FAIL, "performance": DARS.PASS}),
                ("s390x", {"upstream": DARS.PASS, "new": DARS.FAIL}),
            )
        }

        self.assertEqual(
            new_autopkgtest_workflow.output_data,
            OutputData(regression_analysis={}),
        )
        self.schedule_and_run_workflow_callback(new_autopkgtest_workflow)
        self.playground.advance_work_request(
            reference_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        self.schedule_and_run_workflow_callback(new_autopkgtest_workflow)
        self.playground.advance_work_request(
            new_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )

        self.schedule_and_run_workflow_callback(new_autopkgtest_workflow)

        expected_analysis: dict[
            str, tuple[RegressionAnalysisStatus, dict[str, str]]
        ] = {
            "amd64": (
                RegressionAnalysisStatus.STABLE,
                {"upstream": "stable", "old": "stable"},
            )
        }
        new_autopkgtest_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_autopkgtest_workflow,
            "1.0-1",
            reference_results,
            "1.0-2",
            new_results,
            expected_analysis,
        )

        self.assert_callback_status(
            new_autopkgtest_workflow, WorkRequest.Statuses.BLOCKED
        )

        for architecture, reference_result, new_result, status, details in (
            (
                "arm64",
                WorkRequest.Results.SUCCESS,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
                {"upstream": "regression"},
            ),
            (
                "i386",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.IMPROVEMENT,
                {"upstream": "improvement"},
            ),
            (
                "ppc64el",
                None,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.NO_RESULT,
                {"new": "stable"},
            ),
            (
                "riscv64",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.IMPROVEMENT,
                {"upstream": "stable", "performance": "improvement"},
            ),
            (
                "s390x",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
                {"upstream": "improvement", "new": "regression"},
            ),
        ):
            if reference_result is not None:
                self.playground.advance_work_request(
                    reference_tasks[architecture],
                    result=reference_result,
                    use_data=False,
                )
            self.playground.advance_work_request(
                new_tasks[architecture], result=new_result, use_data=False
            )
            expected_analysis[architecture] = (status, details)

        self.assert_callback_status(
            new_autopkgtest_workflow, WorkRequest.Statuses.PENDING
        )

        self.schedule_and_run_workflow_callback(new_autopkgtest_workflow)

        new_autopkgtest_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_autopkgtest_workflow,
            "1.0-1",
            reference_results,
            "1.0-2",
            new_results,
            expected_analysis,
        )
        self.assert_callback_status(
            new_autopkgtest_workflow, WorkRequest.Statuses.COMPLETED
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        context_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )
        self.task_data["context_artifacts"] = [context_artifact.id]
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            AutopkgtestWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-1",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[self.binary_artifact.id],
                context_artifacts_ids=[context_artifact.id],
            ),
        )

    def test_get_input_artifacts(self) -> None:
        context_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )
        self.task_data["context_artifacts"] = [context_artifact.id]
        self.task.dynamic_data = self.task.compute_dynamic_data()
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[self.binary_artifact.id],
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.context_artifacts,
                    label="context_artifacts",
                    artifact_ids=[context_artifact.id],
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Config reader for the debusine clients."""

import logging
import os
import re
import sys
import urllib.parse
from collections.abc import MutableMapping
from configparser import ConfigParser
from pathlib import Path
from typing import NamedTuple, NoReturn, Self, TextIO
from urllib.parse import urlparse

from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT, URLPath, parse_timeout

logger = logging.getLogger(__name__)


class ServerInfo(NamedTuple):
    """Information about a known server."""

    #: Server name to use in configuration
    name: str
    #: User-visible description
    desc: str
    #: Default API URL
    api_url: str
    #: Default Scope
    scope: str
    #: API token
    api_token: str | None = None
    #: Default workspace
    default_workspace: str = "System"
    #: Timeout for HTTP requests (connect, read)
    http_timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT

    def lint(self) -> list[str]:
        """
        Run consistency checks on the current draft configuration.

        :returns: a list of messages, or empty if everything looks ordinary
        """
        res: list[str] = []

        api_url_parsed = urllib.parse.urlparse(self.api_url)

        if api_url_parsed.hostname == "localhost":
            expected_scheme = "http"
        else:
            expected_scheme = "https"

        if api_url_parsed.scheme != expected_scheme:
            res.append(
                f"Server url is an [bold]{api_url_parsed.scheme}[/] url,"
                f" should it be [bold]{expected_scheme}[/]?"
            )

        if not self.api_url.endswith("/api") and not self.api_url.endswith(
            "/api/"
        ):
            res.append("API URL does not end in /api or /api/")

        return res

    @classmethod
    def from_config(
        cls, name: str, section: MutableMapping[str, str], path: Path
    ) -> Self:
        """Infer a ServerInfo from an existing configuration."""
        inferred = cls.from_string(name, desc="temporary serverinfo")
        if known := KNOWN_SERVERS.get(name):
            default_workspace = known.default_workspace
        else:
            default_workspace = "System"
        raw_timeout = section.get("http-timeout")
        timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT

        if raw_timeout:
            try:
                timeout = parse_timeout(raw_timeout)
            except ValueError:
                logger.warning(
                    "Invalid http-timeout value in configuration: %r. "
                    "Falling back to default (%s).",
                    raw_timeout,
                    DEFAULT_REQUEST_TIMEOUT,
                )

        return cls(
            name=name,
            desc=f"Configured in {path}",
            api_url=section.get("api-url", inferred.api_url),
            scope=section.get("scope", inferred.scope),
            api_token=section.get("token", None),
            default_workspace=section.get(
                "default-workspace", default_workspace
            ),
            http_timeout=timeout,
        )

    @classmethod
    def from_string(cls, name: str, desc: str) -> Self | "ServerInfo":
        """Infer a ServerInfo from a user-provided name."""
        if ":" not in name:
            # plain hostname
            server_url_parsed = urllib.parse.ParseResult(
                scheme="",
                netloc=name,
                path="",
                params="",
                query="",
                fragment="",
            )
            server_name = name
        elif re.match(r"^[^/]*:\d+(?:/|$)", name):
            # hostname:port, or just :port
            server_url_parsed = urllib.parse.ParseResult(
                scheme="",
                netloc=(f"localhost{name}" if name.startswith(":") else name),
                path="",
                params="",
                query="",
                fragment="",
            )
            assert server_url_parsed.hostname
            server_name = server_url_parsed.hostname
        else:
            # URL
            server_url_parsed = urllib.parse.urlparse(name)
            assert server_url_parsed.hostname
            server_name = server_url_parsed.hostname

        # Lookup known names
        assert server_url_parsed.hostname
        if known := KNOWN_SERVERS.get(server_url_parsed.hostname):
            return known
        # Lookup known hostnames
        for info in KNOWN_SERVERS.values():
            info_parsed = urllib.parse.urlparse(info.api_url)
            if info_parsed.hostname == server_url_parsed.hostname:
                return info

        if not server_url_parsed.scheme:
            server_url_parsed = server_url_parsed._replace(scheme="https")

        api_path = URLPath(server_url_parsed.path, is_file=False)
        if not api_path or api_path[-1] != "api":
            api_path += "api/"

        api_url = urllib.parse.urlunparse(
            server_url_parsed._replace(path=str(api_path))
        )

        return cls(
            name=server_name, desc=desc, api_url=api_url, scope="debusine"
        )


KNOWN_SERVERS: dict[str, ServerInfo] = {
    info.name: info
    for info in (
        ServerInfo(
            name="debian",
            desc="debusine.debian.net",
            api_url="https://debusine.debian.net/api/",
            scope="debian",
            default_workspace="developers",
        ),
        ServerInfo(
            name="freexian",
            desc="Freexian (E)LTS",
            api_url="https://debusine.freexian.com/api/",
            scope="freexian",
            default_workspace="sandbox",
        ),
        ServerInfo(
            name="localhost",
            desc="debusine-admin runserver for development",
            api_url="http://localhost:8000/api/",
            scope="debusine",
            default_workspace="Playground",
        ),
    )
}


class ConfigHandler(ConfigParser):
    """
    ConfigHandler for the configuration (.ini file) of the Debusine client.

    If the configuration file is not valid it writes an error message
    to stderr and aborts.
    """

    DEFAULT_CONFIG_FILE_PATH = Path.home() / Path(
        '.config/debusine/client/config.ini'
    )

    def __init__(
        self,
        *,
        server_name: str | None = None,
        config_file_path: str | os.PathLike[str] = DEFAULT_CONFIG_FILE_PATH,
        stdout: TextIO = sys.stdout,
        stderr: TextIO = sys.stderr,
    ) -> None:
        """
        Initialize variables and reads the configuration file.

        :param server_name: look up configuration matching this server name
          or FQDN/scope (or None for the default server from the
          configuration)
        :param config_file_path: location of the configuration file
        """
        super().__init__()

        self._server_name = server_name
        self._config_file_path = Path(config_file_path)

        self._stdout = stdout
        self._stderr = stderr

        if not self.read(self._config_file_path):
            self._fail(f'Cannot read {self._config_file_path} .')

    def server_configuration(self) -> ServerInfo:
        """
        Return the server configuration.

        Uses the server specified in the __init__() or the default server
        in the configuration file.
        """
        if self._server_name is None:
            server_name = self._default_server_name()
        else:
            server_name = self._server_name

        return self._server_configuration(server_name)

    def _fail(self, message: str) -> NoReturn:
        """Write message to self._stderr and aborts with exit code 3."""
        self._stderr.write(message + "\n")
        raise SystemExit(3)

    def _default_server_name(self) -> str:
        """Return default server name or aborts."""
        if 'General' not in self:
            self._fail(
                f'[General] section and default-server key must exist '
                f'in {self._config_file_path} to use the default server. '
                f'Add them or specify the server in the command line.'
            )

        if 'default-server' not in self['General']:
            self._fail(
                f'default-server key must exist in [General] in '
                f'{self._config_file_path} . Add it or specify the server '
                f'in the command line.'
            )

        return self['General']['default-server']

    def _server_configuration(self, server_name: str) -> ServerInfo:
        """Return configuration for server_name or aborts."""
        section_name: str | None
        if "/" in server_name:
            # Look up the section by FQDN and scope.
            server_fqdn, scope_name = server_name.split("/")
            for section_name in self.sections():
                if (
                    section_name.startswith("server:")
                    and (
                        (api_url := self[section_name].get("api-url"))
                        is not None
                    )
                    and urlparse(api_url).hostname == server_fqdn
                    and self[section_name].get("scope") == scope_name
                ):
                    break
            else:
                section_name = None
        else:
            # Look up the section by name.
            section_name = f'server:{server_name}'

        if section_name is None or section_name not in self:
            raise ValueError(
                f"No Debusine client configuration for {server_name!r}; "
                f"run 'debusine setup' to configure it"
            )
        server_configuration = self[section_name]

        self._ensure_server_configuration(server_configuration, section_name)

        return ServerInfo.from_config(
            server_name, server_configuration, self._config_file_path
        )

    def _ensure_server_configuration(
        self,
        server_configuration: MutableMapping[str, str],
        server_section_name: str,
    ) -> None:
        """Check keys for server_section_name. Aborts if there are errors."""
        keys = ('api-url', 'scope', 'token')
        optional_keys = frozenset(("default-workspace", "http-timeout"))

        missing_keys = keys - server_configuration.keys()

        if len(missing_keys) > 0:
            self._fail(
                f'Missing required keys in the section '
                f'[{server_section_name}]: {", ".join(sorted(missing_keys))} '
                f'in {self._config_file_path} .'
            )

        extra_keys = server_configuration.keys() - keys - optional_keys

        if len(extra_keys) > 0:
            self._fail(
                f'Invalid keys in the section '
                f'[{server_section_name}]: {", ".join(sorted(extra_keys))} '
                f'in {self._config_file_path} .'
            )



# Generated by Django 4.2.28 on 2026-03-24 14:56

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0047_alter_worker_activation_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='token',
            name='_activating_worker_id',
            field=models.BigIntegerField(
                db_column='activating_worker_id',
                editable=False,
                help_text='Mirror of Worker.activation_token; only for use in constraints',
                null=True,
            ),
        ),
        migrations.AddField(
            model_name='token',
            name='_worker_id',
            field=models.BigIntegerField(
                db_column='worker_id',
                editable=False,
                help_text='Mirror of Worker.token; only for use in constraints',
                null=True,
            ),
        ),
        migrations.AddField(
            model_name='token',
            name='token_type',
            field=models.CharField(
                blank=True,
                choices=[
                    ('worker-activation', 'Worker activation'),
                    ('worker', 'Worker'),
                    ('user', 'User'),
                ],
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='worker',
            trigger=pgtrigger.compiler.Trigger(
                name='db_worker_sync_unset_token',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (OLD."token_id" IS NOT NULL)',
                    func="UPDATE db_token SET worker_id = NULL WHERE db_token.id = OLD.token_id; IF TG_OP = 'DELETE' THEN RETURN OLD; ELSE RETURN NEW; END IF;",
                    hash='f7dcb58150544aad199c222db2908e04c91da36c',
                    operation='UPDATE OF "token_id" OR DELETE',
                    pgid='pgtrigger_db_worker_sync_unset_token_577a2',
                    table='db_worker',
                    when='BEFORE',
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='worker',
            trigger=pgtrigger.compiler.Trigger(
                name='db_worker_sync_unset_activation_token',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (OLD."activation_token_id" IS NOT NULL)',
                    func="UPDATE db_token SET activating_worker_id = NULL WHERE db_token.id = OLD.activation_token_id; IF TG_OP = 'DELETE' THEN RETURN OLD; ELSE RETURN NEW; END IF;",
                    hash='94dbb738ac99d0e3258c71c6dc750ab63cdce39c',
                    operation='UPDATE OF "activation_token_id" OR DELETE',
                    pgid='pgtrigger_db_worker_sync_unset_activation_token_4362c',
                    table='db_worker',
                    when='BEFORE',
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='worker',
            trigger=pgtrigger.compiler.Trigger(
                name='db_worker_sync_set_token',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (NEW."token_id" IS NOT NULL)',
                    func='UPDATE db_token SET worker_id = NEW.id WHERE db_token.id = NEW.token_id; RETURN NEW;',
                    hash='41a37b3ee15b5d073e01597eb49200b33923bb6c',
                    operation='INSERT OR UPDATE OF "token_id"',
                    pgid='pgtrigger_db_worker_sync_set_token_a1f77',
                    table='db_worker',
                    when='AFTER',
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='worker',
            trigger=pgtrigger.compiler.Trigger(
                name='db_worker_sync_set_activation_token',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (NEW."activation_token_id" IS NOT NULL)',
                    func='UPDATE db_token SET activating_worker_id = NEW.id WHERE db_token.id = NEW.activation_token_id; RETURN NEW;',
                    hash='2c21750cfe3932454d16ed0100b9971d4c687d55',
                    operation='INSERT OR UPDATE OF "activation_token_id"',
                    pgid='pgtrigger_db_worker_sync_set_activation_token_dae25',
                    table='db_worker',
                    when='AFTER',
                ),
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the CollectionManagerInterface."""

from typing import override
from unittest.mock import patch

from django.contrib.auth import get_user_model

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
)
from debusine.db.models import Collection, CollectionItem, default_workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections import (
    CollectionManagerInterface,
    DebianSuiteManager,
    ItemAdditionError,
)
from debusine.test.django import TestCase


class DebusineTestManager(CollectionManagerInterface):
    """Collection manager used for testing."""

    COLLECTION_CATEGORY = CollectionCategory.TEST


class CollectionManagerInterfaceTests(TestCase):
    """Tests for CollectionManagerInterface."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.user = get_user_model().objects.create_user(
            username="John", email="john@example.org"
        )
        template = self.playground.create_workflow_template(
            name="test",
            task_name="noop",
        )
        self.workflow = self.playground.create_workflow(
            template, task_data={}, created_by=self.user
        )

        self.workspace = default_workspace()

        self.collection = Collection.objects.create(
            name="Test Collection",
            category=CollectionCategory.TEST,
            workspace=self.workspace,
        )
        self.manager = DebusineTestManager(self.collection)

    def test_add_bare_data(self) -> None:
        """Test add_bare_data() call do_add_bare_data()."""
        category = BareDataCategory.TEST
        self.manager.VALID_BARE_DATA_CATEGORIES = frozenset({category})

        with patch.object(self.manager, "do_add_bare_data") as mocked:
            self.manager.add_bare_data(category, user=self.user)

        mocked.assert_called_with(
            category,
            user=self.user,
            data=None,
            workflow=None,
            name=None,
            replace=False,
            created_at=None,
            replaced_by=None,
        )

    def test_add_bare_data_invalid_category_raise_item_addition_error(
        self,
    ) -> None:
        """Test add_bare_data() raise ItemAdditionError(): invalid category."""
        category = "invalid"

        msg = (
            f'^Bare data category "{category}" not supported by the collection$'
        )
        self.assertRaisesRegex(
            ItemAdditionError,
            msg,
            self.manager.add_bare_data,
            category,
            user=self.user,
        )

    def test_add_bare_data_any_category_valid(self) -> None:
        """If VALID_BARE_DATA_CATEGORIES is None, any category is valid."""
        category = BareDataCategory.TEST
        self.manager.VALID_BARE_DATA_CATEGORIES = None

        with patch.object(self.manager, "do_add_bare_data") as mocked:
            self.manager.add_bare_data(category, user=self.user)

        mocked.assert_called_with(
            category,
            user=self.user,
            workflow=None,
            data=None,
            name=None,
            replace=False,
            created_at=None,
            replaced_by=None,
        )

    def test_add_artifact(self) -> None:
        """Test add_artifact() call do_add_artifact()."""
        category = ArtifactCategory.TEST
        artifact, _ = self.playground.create_artifact(category=category)

        self.manager.VALID_ARTIFACT_CATEGORIES = frozenset({category})

        with patch.object(self.manager, "do_add_artifact") as mocked:
            self.manager.add_artifact(artifact, user=self.user)

        mocked.assert_called_with(
            artifact,
            user=self.user,
            workflow=None,
            variables=None,
            name=None,
            replace=False,
            created_at=None,
            replaced_by=None,
        )

    def test_add_artifact_invalid_category_raise_item_addition_error(
        self,
    ) -> None:
        """Test add_artifact() raise ItemAdditionError(): invalid category."""
        category = "invalid"
        artifact, _ = self.playground.create_artifact(
            category=category,  # type: ignore[arg-type]
        )

        msg = (
            f'^Artifact category "{category}" not supported by the collection$'
        )
        self.assertRaisesRegex(
            ItemAdditionError,
            msg,
            self.manager.add_artifact,
            artifact,
            user=self.user,
        )

    def test_add_artifact_any_category_valid(self) -> None:
        """If VALID_ARTIFACT_CATEGORIES is None, any category is valid."""
        category = ArtifactCategory.TEST
        artifact, _ = self.playground.create_artifact(category=category)
        self.manager.VALID_ARTIFACT_CATEGORIES = None

        with patch.object(self.manager, "do_add_artifact") as mocked:
            self.manager.add_artifact(artifact, user=self.user)

        mocked.assert_called_with(
            artifact,
            user=self.user,
            workflow=None,
            variables=None,
            name=None,
            replace=False,
            created_at=None,
            replaced_by=None,
        )

    def test_add_collection(self) -> None:
        """Test add_collection() call do_add_collection()."""
        category = CollectionCategory.TEST

        self.manager.VALID_COLLECTION_CATEGORIES = frozenset({category})

        collection = Collection.objects.create(
            name="Testing", category=category, workspace=self.workspace
        )

        with patch.object(self.manager, "do_add_collection") as mocked:
            self.manager.add_collection(collection, user=self.user)

        mocked.assert_called_with(
            collection,
            user=self.user,
            workflow=None,
            variables=None,
            name=None,
            replace=False,
            created_at=None,
            replaced_by=None,
        )

    def test_add_collection_invalid_category_raise_item_addition_error(
        self,
    ) -> None:
        """Test add_artifact() raise ItemAdditionError(): invalid category."""
        category = "invalid"
        collection = Collection.objects.create(
            name="Some name", category=category, workspace=self.workspace
        )

        msg = (
            f'^Collection category "{category}" '
            f'not supported by the collection$'
        )
        self.assertRaisesRegex(
            ItemAdditionError,
            msg,
            self.manager.add_collection,
            collection,
            user=self.user,
        )

    def test_add_collection_any_category_valid(self) -> None:
        """If VALID_COLLECTION_CATEGORIES is None, any category is valid."""
        category = "debusine:anything"
        collection = Collection.objects.create(
            name="Some name", category=category, workspace=self.workspace
        )
        self.manager.VALID_COLLECTION_CATEGORIES = None

        with patch.object(self.manager, "do_add_collection") as mocked:
            self.manager.add_collection(collection, user=self.user)

        mocked.assert_called_with(
            collection,
            user=self.user,
            workflow=None,
            variables=None,
            name=None,
            replace=False,
            created_at=None,
            replaced_by=None,
        )

    def test_remove_item_bare_data(self) -> None:
        """remove_item() calls do_remove_item() for bare data."""
        item = self.playground.create_bare_data_item(self.collection, "test")

        with patch.object(self.manager, "do_remove_item") as mocked:
            self.manager.remove_item(
                item, user=self.user, workflow=self.workflow
            )

        mocked.assert_called_with(item, user=self.user, workflow=self.workflow)

    def test_remove_item_artifact(self) -> None:
        """remove_item() calls do_remove_item() for an artifact."""
        artifact, _ = self.playground.create_artifact()
        item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.playground.get_default_user(),
        )

        with patch.object(self.manager, "do_remove_item") as mocked:
            self.manager.remove_item(item, user=self.user, workflow=None)

        mocked.assert_called_with(item, user=self.user, workflow=None)

    def test_remove_item_collection(self) -> None:
        """remove_item() calls do_remove_item() for a collection."""
        collection = Collection.objects.create(
            name="Testing", category="Testing", workspace=self.workspace
        )
        item = CollectionItem.objects.create_from_collection(
            collection,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.playground.get_default_user(),
        )

        with patch.object(self.manager, "do_remove_item") as mocked:
            self.manager.remove_item(
                item, user=self.user, workflow=self.workflow
            )

        mocked.assert_called_with(item, user=self.user, workflow=self.workflow)

    def test_remove_items_by_name_bare_data(self) -> None:
        """remove_items_by_name() calls do_remove_item() for bare data."""
        item = self.playground.create_bare_data_item(self.collection, "test")

        with patch.object(self.manager, "do_remove_item") as mocked:
            self.manager.remove_items_by_name(
                name="test",
                child_types=[CollectionItem.Types.BARE],
                user=self.user,
                workflow=self.workflow,
            )

        mocked.assert_called_with(item, user=self.user, workflow=self.workflow)

    def test_remove_items_by_name_artifact(self) -> None:
        """remove_items_by_name() calls do_remove_item() for an artifact."""
        artifact, _ = self.playground.create_artifact()
        item = CollectionItem.objects.create(
            name="test",
            artifact=artifact,
            parent_collection=self.collection,
            child_type=CollectionItem.Types.ARTIFACT,
            created_by_user=self.user,
        )

        with patch.object(self.manager, "do_remove_item") as mocked:
            self.manager.remove_items_by_name(
                name="test",
                child_types=[CollectionItem.Types.ARTIFACT],
                user=self.user,
                workflow=self.workflow,
            )

        mocked.assert_called_with(item, user=self.user, workflow=self.workflow)

    def test_remove_items_by_name_collection(self) -> None:
        """remove_items_by_name() calls do_remove_item() for a collection."""
        collection = self.playground.create_collection(
            name="Collection Test", category=CollectionCategory.TEST
        )

        item = CollectionItem.objects.create(
            name="test",
            collection=collection,
            parent_collection=self.collection,
            child_type=CollectionItem.Types.COLLECTION,
            created_by_user=self.user,
        )

        with patch.object(self.manager, "do_remove_item") as mocked:
            self.manager.remove_items_by_name(
                name="test",
                child_types=[CollectionItem.Types.COLLECTION],
                user=self.user,
                workflow=self.workflow,
            )

        mocked.assert_called_with(item, user=self.user, workflow=self.workflow)

    def test_lookup_unexpected_format_raise_lookup_error(self) -> None:
        """`lookup` raises `LookupError`: invalid format."""
        with self.assertRaisesRegex(
            LookupError, '^Unexpected lookup format: "nonsense"$'
        ):
            self.manager.lookup("nonsense")

    def test_lookup_return_none(self) -> None:
        """`lookup` returns None if there are no matches."""
        self.assertIsNone(self.manager.lookup("name:nonexistent"))

    def test_lookup_filter_unexpected_format_raise_lookup_error(self) -> None:
        """`lookup_filter` raises `LookupError`: invalid format."""
        with self.assertRaisesRegex(
            LookupError, '^Unexpected lookup filter format: "nonsense"$'
        ):
            self.manager.lookup_filter(
                "nonsense",
                "foo",
                workspace=self.playground.get_default_workspace(),
                pc=PermissionContext.for_ui_user(
                    self.playground.get_default_user()
                ),
            )

    def test_get_manager_for(self) -> None:
        """Test getting specialized manager."""
        collection = Collection.objects.create(
            name="Testing",
            category=CollectionCategory.SUITE,
            workspace=self.workspace,
        )
        manager = CollectionManagerInterface.get_manager_for(collection)
        self.assertEqual(type(manager), DebianSuiteManager)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the QemuExecutor class."""

from pathlib import Path
from typing import override
from unittest import mock
from unittest.mock import MagicMock

from debusine.client.debusine import Debusine
from debusine.tasks import Noop
from debusine.tasks.executors.base import (
    ExecutorStatistics,
    ImageImportError,
    ImageNotDownloadedError,
    executor_class,
)
from debusine.tasks.executors.qemu import QemuExecutor
from debusine.tasks.models import BackendType
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase


class QemuExecutorTests(ExternalTaskHelperMixin[Noop], TestCase):
    """Unit tests for QemuExecutor."""

    @override
    def setUp(self) -> None:
        """Mock the Debusine API for tests."""
        super().setUp()
        self.debusine_api = MagicMock(spec=Debusine)
        self.image_artifact = self.mock_image_download(
            self.debusine_api, system_image=True
        )
        self.image_artifact.data["pkglist"]["python3-minimal"] = "3.0"

        self.workdir = self.create_temporary_directory()
        self.executor = QemuExecutor(
            self.debusine_api,
            42,
            workspace="workspace",
            working_directory=self.workdir,
        )

    def test_backend_name(self) -> None:
        """Test that the backend_name attribute was set."""
        self.assertEqual(QemuExecutor.backend_name, BackendType.QEMU)

    @mock.patch("platform.machine")
    @mock.patch("os.access")
    def test_available(self, access: MagicMock, machine: MagicMock) -> None:
        """available() returns True if qemu and /dev/kvm are available."""
        access.return_value = True
        machine.return_value = "x86_64"
        self.mock_is_command_available({"qemu-system-x86_64": True})
        self.assertTrue(QemuExecutor.available())

    @mock.patch("platform.machine")
    @mock.patch("os.access")
    def test_available_no_qemu(
        self, access: MagicMock, machine: MagicMock
    ) -> None:
        """Test that available() returns False if qemu is not available."""
        access.return_value = True
        machine.return_value = "x86_64"
        self.mock_is_command_available({"qemu-system-x86_64": False})
        self.assertFalse(QemuExecutor.available())

    @mock.patch("platform.machine")
    @mock.patch("os.access")
    def test_available_no_kvm(
        self, access: MagicMock, machine: MagicMock
    ) -> None:
        """Test that available() returns False if kvm is not available."""
        access.return_value = False
        machine.return_value = "x86_64"
        self.mock_is_command_available({"qemu-system-x86_64": True})
        self.assertFalse(QemuExecutor.available())

    def test_instantiation_fetches_artifact(self) -> None:
        """Test that instantiating QemuExecutor fetches the artifact."""
        self.assertEqual(self.executor.system_image, self.image_artifact)

    def test_download_image(self) -> None:
        """Test that download_image calls ImageCache.download_image."""
        response = self.executor.download_image()
        expected_path = self.image_cache_path / "42/image.qcow2"
        self.assertEqual(self.executor._local_path, expected_path)
        self.assertEqual(response, str(expected_path))

    def test_download_image_requires_python3(self) -> None:
        """Test that download_image requires python3-minimal in the image."""
        del self.image_artifact.data["pkglist"]["python3-minimal"]
        with self.assertRaisesRegex(
            ImageImportError,
            (
                "Image doesn't contain a python3-minimal, required by qemu "
                "backend."
            ),
        ):
            self.executor.download_image()

    def test_image_name(self) -> None:
        """Test that image_name returns the previously downloaded image path."""
        self.executor.download_image()
        response = self.executor.image_name()
        expected_path = self.image_cache_path / "42/image.qcow2"
        self.assertEqual(response, str(expected_path))

    def test_image_name_not_downloaded(self) -> None:
        """Test that image_name fails if there is no name."""
        with self.assertRaisesRegex(
            ImageNotDownloadedError, r"download_image\(\) not yet called"
        ):
            self.executor.image_name()

    def test_autopkgtest_virt_server(self) -> None:
        """Test that autopkgtest_virt_server returns qemu."""
        self.assertEqual(self.executor.autopkgtest_virt_server(), "qemu")

    def test_autopkgtest_virt_args(self) -> None:
        """Test that autopkgtest_virt_args returns sane arguments."""
        self.executor._local_path = Path("/not/used/path/image.qcow2")
        self.assertEqual(
            self.executor.autopkgtest_virt_args(),
            [
                "--boot",
                "efi",
                "--dpkg-architecture",
                "amd64",
                str(self.executor._local_path),
            ],
        )

    def test_autopkgtest_virt_args_before_download(self) -> None:
        """Test that autopkgtest_virt_args requires download."""
        with self.assertRaisesRegex(
            ImageNotDownloadedError, r"download_image\(\) not yet called"
        ):
            self.executor.autopkgtest_virt_args()

    def test_executor_class_finds_qemu(self) -> None:
        """Test that executor_class() supports qemu."""
        instance = executor_class(BackendType.QEMU)
        self.assertEqual(instance, QemuExecutor)

    def test_create(self) -> None:
        """Test create() raises NotImplemented."""
        with self.assertRaises(NotImplementedError):
            self.executor.create()

    def test_get_statistics(self) -> None:
        """`get_statistics` returns empty statistics."""
        self.assertEqual(
            QemuExecutor.get_statistics(None), ExecutorStatistics()
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the TokenLastSeenAtMiddleware."""

from collections.abc import Mapping
from typing import Any, TYPE_CHECKING, cast

from asgiref.typing import (
    ASGIReceiveCallable,
    ASGISendCallable,
    WebSocketReceiveEvent,
)
from channels.db import database_sync_to_async
from django.utils import timezone

from debusine.db.context import context
from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.server.consumers import ChannelScopeWithToken
from debusine.server.middlewares.token_last_seen_at import (
    TokenLastSeenAtMiddlewareChannels,
)
from debusine.test.django import TransactionTestCase

if TYPE_CHECKING:
    from channels.consumer import _ChannelScope


class TokenLastSeenAtMiddlewareChannelsTests(TransactionTestCase):
    """Test TokenLastSeenAtMiddlewareChannels."""

    scenario = scenarios.DefaultScopeUser()

    @classmethod
    async def _noop_receive(cls) -> WebSocketReceiveEvent:  # pragma: no cover
        """No-op receive callback for testing."""
        return WebSocketReceiveEvent(
            type="websocket.receive", bytes=None, text=""
        )

    @classmethod
    async def _noop_send(cls, message: Mapping[str, Any]) -> None:
        """No-op send callback for testing."""

    async def test_update_last_seen_at_store_user_token(self) -> None:
        """Request with a token updates token's last_seen_at."""
        token = await database_sync_to_async(
            self.playground.create_user_token
        )()
        assert token is not None

        self.assertIsNone(token.last_seen_at)

        before = timezone.now()

        scope_upstream = cast(
            "_ChannelScope",
            {
                "type": "websocket",
                "headers": [(b"token", token.key.encode("utf-8"))],
            },
        )

        async def mock_app(
            scope_downstream: ChannelScopeWithToken,
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            """Mock ASGI app to inspect the modified scope."""
            await token.arefresh_from_db()
            self.assertEqual(scope_downstream["token"], token)
            self.assertIsNone(context._permission_context.get())
            # self.assertEqual(context.worker_token, token)
            assert token.last_seen_at is not None
            self.assertGreaterEqual(token.last_seen_at, before)
            self.assertLess(token.last_seen_at, timezone.now())

        middleware = TokenLastSeenAtMiddlewareChannels(app=mock_app)

        await middleware(scope_upstream, self._noop_receive, self._noop_send)

    async def test_update_last_seen_at_store_worker_token(self) -> None:
        """Request with a token updates token's last_seen_at."""
        token = await database_sync_to_async(
            self.playground.create_worker_token
        )()
        assert token is not None

        self.assertIsNone(token.last_seen_at)

        before = timezone.now()

        scope_upstream = cast(
            "_ChannelScope",
            {
                "type": "websocket",
                "headers": [(b"token", token.key.encode("utf-8"))],
            },
        )

        async def mock_app(
            scope_downstream: ChannelScopeWithToken,
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            """Mock ASGI app to inspect the modified scope."""
            await token.arefresh_from_db()
            self.assertEqual(scope_downstream["token"], token)
            self.assertEqual(context.pc.worker_token, token)
            assert token.last_seen_at is not None
            self.assertGreaterEqual(token.last_seen_at, before)
            self.assertLess(token.last_seen_at, timezone.now())

        middleware = TokenLastSeenAtMiddlewareChannels(app=mock_app)

        await middleware(scope_upstream, self._noop_receive, self._noop_send)

    async def test_update_last_seen_at_store_work_request_token(self) -> None:
        """Request with a work request token updates token's last_seen_at."""
        token = await database_sync_to_async(
            self.playground.create_work_request_token
        )()
        assert token is not None

        self.assertIsNone(token.last_seen_at)

        before = timezone.now()

        scope_upstream = cast(
            "_ChannelScope",
            {
                "type": "websocket",
                "headers": [(b"token", token.key.encode("utf-8"))],
            },
        )

        async def mock_app(
            scope_downstream: ChannelScopeWithToken,
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            """Mock ASGI app to inspect the modified scope."""
            await token.arefresh_from_db()
            self.assertEqual(scope_downstream["token"], token)
            self.assertEqual(context.pc.work_request_token, token)
            assert token.last_seen_at is not None
            self.assertGreaterEqual(token.last_seen_at, before)
            self.assertLess(token.last_seen_at, timezone.now())

        middleware = TokenLastSeenAtMiddlewareChannels(app=mock_app)

        await middleware(scope_upstream, self._noop_receive, self._noop_send)

    async def test_request_without_token(self) -> None:
        """Request without a token."""
        scope_upstream = cast(
            "_ChannelScope",
            {"type": "websocket", "headers": [(b"user-agent", b"foo")]},
        )

        async def mock_app(
            scope_downstream: ChannelScopeWithToken,
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            """Mock ASGI app to inspect the modified scope."""
            self.assertIsNone(scope_downstream["token"])
            self.assertIsNone(context._permission_context.get())

        middleware = TokenLastSeenAtMiddlewareChannels(app=mock_app)

        await middleware(scope_upstream, self._noop_receive, self._noop_send)

    async def test_request_token_key_does_not_exist(self) -> None:
        """Request where the Token key in the header does not exist."""
        scope_upstream = cast(
            "_ChannelScope",
            {"type": "websocket", "headers": [(b"token", b"DOES-NOT-EXIST")]},
        )

        async def mock_app(
            scope_downstream: ChannelScopeWithToken,
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            """Mock ASGI app to inspect the modified scope."""
            self.assertFalse(hasattr(scope_downstream, "token"))
            self.assertIsNone(context._permission_context.get())

        middleware = TokenLastSeenAtMiddlewareChannels(app=mock_app)
        await middleware(scope_upstream, self._noop_receive, self._noop_send)

    async def test_request_token_disabled(self) -> None:
        """Request with a disabled Token."""
        token = await database_sync_to_async(
            self.playground.create_worker_token
        )(enabled=False)

        scope_upstream = cast(
            "_ChannelScope",
            {
                "type": "websocket",
                "headers": [(b"token", token.key.encode("utf-8"))],
            },
        )

        async def mock_app(
            scope_downstream: ChannelScopeWithToken,
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            """Mock ASGI app to inspect the modified scope."""
            self.assertEqual(scope_downstream["token"], token)
            self.assertIsNone(context._permission_context.get())

        middleware = TokenLastSeenAtMiddlewareChannels(app=mock_app)
        await middleware(scope_upstream, self._noop_receive, self._noop_send)

    async def test_work_request_token_invalid_state(self) -> None:
        token = await database_sync_to_async(
            self.playground.create_work_request_token
        )()
        assert token is not None

        def set_blocked() -> None:
            assert token.work_request is not None
            token.work_request.status = WorkRequest.Statuses.BLOCKED
            token.work_request.save()

        await database_sync_to_async(set_blocked)()

        self.assertIsNone(token.last_seen_at)

        before = timezone.now()

        scope_upstream = cast(
            "_ChannelScope",
            {
                "type": "websocket",
                "headers": [(b"token", token.key.encode("utf-8"))],
            },
        )

        async def mock_app(
            scope_downstream: ChannelScopeWithToken,
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            """Mock ASGI app to inspect the modified scope."""
            await token.arefresh_from_db()
            self.assertEqual(scope_downstream["token"], token)
            self.assertIsNone(context._permission_context.get())
            assert token.last_seen_at is not None
            self.assertGreaterEqual(token.last_seen_at, before)
            self.assertLess(token.last_seen_at, timezone.now())

        middleware = TokenLastSeenAtMiddlewareChannels(app=mock_app)

        await middleware(scope_upstream, self._noop_receive, self._noop_send)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: rest framework extensions."""

from collections.abc import Iterable
from typing import Any, TYPE_CHECKING, override

from django.http import JsonResponse
from rest_framework import status
from rest_framework.permissions import BasePermission
from rest_framework.request import Request

from debusine.db.models import Token

if TYPE_CHECKING:
    from rest_framework.views import APIView


class ProblemResponse(JsonResponse):
    """
    Holds a title and other optional fields to return problems to the client.

    Follows RFC7807 (https://www.rfc-editor.org/rfc/rfc7807#section-6.1)
    """

    def __init__(
        self,
        title: str,
        detail: str | None = None,
        validation_errors: Iterable[Any] | None = None,
        status_code: int = status.HTTP_400_BAD_REQUEST,
    ) -> None:
        """
        Initialize object.

        :param title: included in the response data.
        :param detail: if not None, included in the response data.
        :param validation_errors: if not None, included in the response data.
        :param status_code: HTTP status code for the response.
        """
        data: dict[str, Any] = {"title": title}

        if detail is not None:
            data["detail"] = detail

        if validation_errors is not None:
            data["validation_errors"] = validation_errors

        super().__init__(
            data, status=status_code, content_type="application/problem+json"
        )


class IsTokenAuthenticated(BasePermission):
    """Allow access to requests with a valid token."""

    @override
    def has_permission(
        self,
        request: Request,
        view: "APIView",  # noqa: ARG002, U100
    ) -> bool:
        """Return True if the request is authenticated with a Token."""
        from debusine.db.models import Token

        return isinstance(request.auth, Token) and request.auth.enabled


class IsUserTokenAuthenticated(IsTokenAuthenticated):
    """Allow access if the request has an enabled Token with associated user."""

    @override
    def has_permission(
        self,
        request: Request,
        view: "APIView",  # noqa: ARG002, U100
    ) -> bool:
        """Return True if valid token has a User assigned."""
        from debusine.db.models import Token

        return (
            isinstance(request.auth, Token)
            and request.auth.enabled
            and request.auth.token_type == Token.Types.USER
        )


class IsWorkerTokenAuthenticated(IsTokenAuthenticated):
    """Allow access to requests with a token assigned to a worker."""

    @override
    def has_permission(self, request: Request, view: "APIView") -> bool:
        """
        Return True if the request is authenticated to an external worker.

        The Token must exist in the database and have a Worker.
        """
        if not super().has_permission(request, view):
            # No token authenticated: no Worker Authenticated
            return False

        if (
            not isinstance(request.auth, Token)
            or request.auth.token_type != Token.Types.WORKER
        ):
            # The request is not authenticated by a worker token.
            return False

        return True


class IsWorkRequestTokenAuthenticated(IsTokenAuthenticated):
    """Allow access to an external worker to complete a work request."""

    @override
    def has_permission(self, request: Request, view: "APIView") -> bool:
        """
        Return True if the request is authenticated with a work request token.

        The Token must exist in the database and have a WorkRequest assigned.
        """
        if not super().has_permission(request, view):
            return False

        if (
            not isinstance(request.auth, Token)
            or request.auth.token_type != Token.Types.WORK_REQUEST
        ):
            return False

        return True


class IsGet(BasePermission):
    """Allow access if the request's method is GET."""

    @override
    def has_permission(
        self,
        request: Request,
        view: "APIView",  # noqa: ARG002, U100
    ) -> bool:
        """Return True if request.method == "GET"."""
        return request.method == "GET"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command test."""

import copy
import os.path
import shutil
from typing import Any, override
from unittest import mock

from django.conf import settings

from debusine.django.django_utils import setup_temp_data_directories
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


class TestCommandTests(TestCase):
    """Tests for test command."""

    @override
    def setUp(self) -> None:
        """Save settings."""
        super().setUp()
        self.original_settings: dict[str, Any] = {
            "DEBUSINE_DATA_PATH": "",
            "DEBUSINE_CACHE_DIRECTORY": "",
            "DEBUSINE_TEMPLATE_DIRECTORY": "",
            "DEBUSINE_UPLOAD_DIRECTORY": "",
            "DEBUSINE_STORE_DIRECTORY": "",
            "DEBUSINE_TOKENS_DIRECTORY": "",
            "TEMPLATES": None,
        }
        for key in self.original_settings:
            self.original_settings[key] = copy.deepcopy(getattr(settings, key))

    @override
    def tearDown(self) -> None:
        """Restore settings."""
        for key in self.original_settings:
            setattr(settings, key, self.original_settings[key])
        super().tearDown()

    def test_data_directory_removal(self) -> None:
        """Test '--keepdata'."""
        with mock.patch(
            "django.core.management.commands.test.Command.handle", autospec=True
        ):
            setup_temp_data_directories()
            call_command("test")
            self.assertFalse(os.path.exists(settings.DEBUSINE_DATA_PATH))

            setup_temp_data_directories()
            call_command("test", keepdata=True, verbosity=0)
            self.assertTrue(os.path.exists(settings.DEBUSINE_DATA_PATH))
            shutil.rmtree(settings.DEBUSINE_DATA_PATH)

            setup_temp_data_directories()
            call_command("test", keepdata=True, verbosity=1)
            self.assertTrue(os.path.exists(settings.DEBUSINE_DATA_PATH))
            shutil.rmtree(settings.DEBUSINE_DATA_PATH)

    def test_template_directory(self) -> None:
        self.assertEqual(
            settings.TEMPLATES[0]["DIRS"],
            [settings.DEBUSINE_TEMPLATE_DIRECTORY],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Signing task to generate a new key."""

import base64
from pathlib import Path
from typing import override

from debusine.artifacts.models import WorkRequestResults
from debusine.assets import AssetCategory, SigningKeyData
from debusine.signing.db.models import Key
from debusine.signing.tasks import BaseSigningTask
from debusine.signing.tasks.models import GenerateKeyData


class GenerateKey(BaseSigningTask[GenerateKeyData]):
    """Task that generates a new key."""

    TASK_VERSION = 1

    _key: Key | None = None

    @override
    def get_subject(self) -> str | None:
        return None

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
        """Execute the task."""
        assert self.work_request_id is not None
        assert self.debusine is not None

        with self.open_debug_log_file("cmd-output.log", mode="wb") as log_file:
            self._key = Key.objects.generate(
                Key.Purpose(self.data.purpose),
                self.data.description,
                self.work_request_id,
                log_file=log_file,
            )

        return WorkRequestResults.SUCCESS

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,
    ) -> None:
        """Create a signing key asset."""
        assert self.debusine is not None
        assert self.workspace_name is not None
        # run() always either returns SUCCESS or raises an exception; in the
        # latter case we won't get here.
        assert execution_result == WorkRequestResults.SUCCESS
        assert self._key is not None

        self.debusine.asset_create(
            category=AssetCategory.SIGNING_KEY,
            data=SigningKeyData(
                purpose=self.data.purpose,
                fingerprint=self._key.fingerprint,
                public_key=base64.b64encode(self._key.public_key).decode(),
                description=self.data.description,
            ),
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )



#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine integration tests for APT mirroring."""

import lzma
import os
import re
import socket
import subprocess
import unittest
from functools import partial
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path
from threading import Thread
from typing import override

import gpg
import requests
import yaml

from debusine.signing.gnupg import gpg_ephemeral_context, gpg_generate
from utils.client import Client
from utils.common import Configuration
from utils.integration_test_helpers_mixin import IntegrationTestHelpersMixin
from utils.server import DebusineServer


class IntegrationTaskAPTMirrorTests(
    IntegrationTestHelpersMixin, unittest.TestCase
):
    """
    Integration tests for the aptmirror task.

    These tests assume:
    - debusine-server is running
    - debusine-worker is running (connected to the server)
    - debusine-client is correctly configured
    """

    TASK_NAME = "aptmirror"

    @override
    def setUp(self) -> None:
        super().setUp()
        # If debusine-server or nginx was launched just before the
        # integration-tests-simplesystemimagebuild.py is launched the
        # debusine-server might not be yet available. Let's wait for the
        # debusine-server to be reachable if it's not ready yet
        self.assertTrue(
            DebusineServer.wait_for_server_ready(),
            f"debusine-server should be available (in "
            f"{Configuration.get_base_url()}) before the integration tests "
            f"are run",
        )

        self.architecture = subprocess.check_output(
            ["dpkg", "--print-architecture"], text=True
        ).strip()

    def make_packages(
        self,
        temp_path: Path,
        packages_path: Path,
        from_directory: str,
        override_path: Path,
        architecture: str | None = None,
    ) -> None:
        """Generate ``Packages`` and ``Packages.xz`` files."""
        args: list[str | os.PathLike[str]] = ["apt-ftparchive"]
        if architecture is not None:
            args += ["-a", architecture]
        args += ["packages", from_directory, override_path]
        packages_data = subprocess.run(
            args, stdout=subprocess.PIPE, cwd=temp_path, check=True
        ).stdout
        with open(packages_path, mode="wb") as packages_file:
            packages_file.write(packages_data)
        with lzma.open(
            packages_path.with_suffix(".xz"), mode="wb", format=lzma.FORMAT_XZ
        ) as packages_xz_file:
            packages_xz_file.write(packages_data)

    def make_sources(
        self,
        temp_path: Path,
        sources_path: Path,
        from_directory: str,
        override_path: Path,
    ) -> None:
        """Generate ``Sources`` and ``Sources.xz`` files."""
        sources_data = subprocess.run(
            ["apt-ftparchive", "sources", from_directory, override_path],
            stdout=subprocess.PIPE,
            cwd=temp_path,
            check=True,
        ).stdout
        with open(sources_path, mode="wb") as sources_file:
            sources_file.write(sources_data)
        with lzma.open(
            sources_path.with_suffix(".xz"), mode="wb", format=lzma.FORMAT_XZ
        ) as sources_xz_file:
            sources_xz_file.write(sources_data)

    def make_release(self, release_path: Path, suite_name: str) -> None:
        """Generate a ``Release`` file."""
        with open(release_path, "w") as release:
            subprocess.run(
                [
                    "apt-ftparchive",
                    "-o",
                    f"APT::FTPArchive::Release::Suite={suite_name}",
                    "release",
                    release_path.parent,
                ],
                stdout=release,
                text=True,
                check=True,
            )

    def gpg_sign(
        self, private_key: bytes, data_path: Path, signature_path: Path
    ) -> None:
        """
        Sign data using GnuPG.

        Borrowed from ``debusine.signing.gnupg.gpg_sign``, but without the
        ``ProtectedKey`` abstraction that just gets in the way for
        integration testing.  Don't use this in production.
        """
        with (
            self._temporary_directory() as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            import_result = ctx.key_import(private_key)
            self.assertNotEqual(getattr(import_result, "secret_imported", 0), 0)
            ctx.signers = list(
                ctx.keylist(pattern=import_result.imports[0].fpr, secret=True)
            )
            with (
                open(data_path, mode="rb") as data_file,
                open(signature_path, mode="wb") as signature_file,
            ):
                ctx.sign(
                    gpg.Data(file=data_file),
                    sink=signature_file,
                    mode=gpg.constants.SIG_MODE_CLEAR,
                )

    def serve_archive(self, path: Path) -> tuple[str, int]:
        """Serve a local archive over HTTP, in a thread."""
        httpd = HTTPServer(
            ("", 0),
            # TODO: Simplify to `directory=path` once our mypy includes
            # https://github.com/python/typeshed/pull/13477
            partial(SimpleHTTPRequestHandler, directory=path.as_posix()),
        )
        httpd_thread = Thread(target=httpd.serve_forever)
        httpd_thread.start()
        self.addCleanup(httpd_thread.join)
        self.addCleanup(httpd.shutdown)
        host, port, *_ = httpd.server_address
        assert isinstance(host, str)
        return host, port

    def wait_and_assert_result(self, work_request_id: int) -> None:
        """Wait for a task to complete, assert success."""
        status = Client.wait_for_work_request_completed(
            work_request_id, "success"
        )
        if not status:
            self.print_work_request_debug_logs(work_request_id)
        self.assertTrue(status)

    def test_pooled(self) -> None:
        """Import and serve packages from a pooled format."""
        suite_name = "sid-dists-pool"
        architectures = ["amd64", "s390x"]
        self.create_suite(suite_name, architectures=architectures)
        self.add_suite_to_archive(suite_name)

        private_key, public_key = gpg_generate("test")

        with (
            self.apt_indexes("sid", architectures=architectures) as apt_path,
            self._temporary_directory() as temp_path,
        ):
            debusine_path = temp_path / "pool/main/d/debusine"
            debusine_path.mkdir(parents=True)
            hello_path = temp_path / "pool/main/h/hello"
            hello_path.mkdir(parents=True)
            subprocess.run(
                ["apt-get", "source", "--download-only", "debusine"],
                cwd=debusine_path,
                env=self.make_apt_environment(apt_path),
                check=True,
            )
            for package, parent_path in (
                ("python3-debusine", debusine_path),
                ("hello:amd64", hello_path),
                ("hello:s390x", hello_path),
            ):
                subprocess.run(
                    ["apt-get", "download", package],
                    cwd=parent_path,
                    env=self.make_apt_environment(apt_path),
                    check=True,
                )
            suite_path = temp_path / f"dists/{suite_name}"
            suite_binary_amd64_path = suite_path / "main/binary-amd64"
            suite_binary_s390x_path = suite_path / "main/binary-s390x"
            suite_source_path = suite_path / "main/source"
            suite_binary_amd64_path.mkdir(parents=True)
            suite_binary_s390x_path.mkdir(parents=True)
            suite_source_path.mkdir(parents=True)
            override_path = suite_path / "override"
            override_path.write_text(
                "hello optional devel\npython3-debusine optional python\n"
            )
            override_path.with_suffix(".src").write_text("debusine misc\n")
            for path, architecture in (
                (suite_binary_amd64_path, "amd64"),
                (suite_binary_s390x_path, "s390x"),
            ):
                self.make_packages(
                    temp_path,
                    path / "Packages",
                    "pool",
                    override_path,
                    architecture=architecture,
                )
            self.make_sources(
                temp_path, suite_source_path / "Sources", "pool", override_path
            )
            self.make_release(suite_path / "Release", suite_name)
            (suite_binary_amd64_path / "Packages").unlink()
            (suite_source_path / "Sources").unlink()
            self.gpg_sign(
                private_key, suite_path / "Release", suite_path / "InRelease"
            )
            pool_files = {
                # Strip the version.  This isn't quite accurate, but it's
                # fine as long as we only care about .dsc and .deb files.
                re.sub(r"_[^_]+([_.])", r"\1", filename): (
                    filename,
                    (dirpath / filename).read_bytes(),
                )
                for dirpath, _, filenames in (temp_path / "pool").walk()
                for filename in filenames
            }
            host, port = self.serve_archive(temp_path)

            result = DebusineServer.execute_command(
                "create_work_request",
                "--created-by",
                "test-user",
                "server",
                "aptmirror",
                stdin=yaml.safe_dump(
                    {
                        "collection": suite_name,
                        "url": f"http://{host}:{port}/",
                        "suite": suite_name,
                        "components": ["main"],
                        "architectures": architectures,
                        "signing_key": public_key.decode(),
                    }
                ),
            )
            self.assertEqual(result.returncode, 0)
            self.wait_and_assert_result(
                yaml.safe_load(result.stdout)["work_request_id"]
            )

        archive_url = f"https://deb.{socket.getfqdn()}/debusine/System"
        with (
            self.apt_indexes(
                suite_name,
                url=archive_url,
                signed_by=public_key.decode(),
                architectures=architectures,
            ) as apt_path,
            self._temporary_directory() as temp_path,
        ):
            subprocess.run(
                ["apt-get", "source", "--download-only", "debusine"],
                cwd=temp_path,
                env=self.make_apt_environment(apt_path),
                check=True,
            )
            self.assertEqual(
                (temp_path / pool_files["debusine.dsc"][0]).read_bytes(),
                pool_files["debusine.dsc"][1],
            )
            for package, architecture in (
                ("python3-debusine", "all"),
                ("hello", "amd64"),
                ("hello", "s390x"),
            ):
                subprocess.run(
                    ["apt-get", "download", f"{package}:{architecture}"],
                    cwd=temp_path,
                    env=self.make_apt_environment(apt_path),
                    check=True,
                )
                self.assertEqual(
                    (
                        temp_path
                        / pool_files[f"{package}_{architecture}.deb"][0]
                    ).read_bytes(),
                    pool_files[f"{package}_{architecture}.deb"][1],
                )

    def test_flat(self) -> None:
        """Import and serve packages from a flat format."""
        suite_name = "sid-flat"
        self.create_suite(suite_name, architectures=[self.architecture])
        self.add_suite_to_archive(suite_name)

        private_key, public_key = gpg_generate("test")

        with (
            self.apt_indexes("sid") as apt_path,
            self._temporary_directory() as temp_path,
        ):
            subprocess.run(
                ["apt-get", "source", "--download-only", "debusine"],
                cwd=temp_path,
                env=self.make_apt_environment(apt_path),
                check=True,
            )
            subprocess.run(
                ["apt-get", "download", "python3-debusine"],
                cwd=temp_path,
                env=self.make_apt_environment(apt_path),
                check=True,
            )
            override_path = temp_path / "override"
            override_path.write_text("python3-debusine optional python\n")
            override_path.with_suffix(".src").write_text("debusine misc\n")
            self.make_packages(
                temp_path, temp_path / "Packages", ".", override_path
            )
            self.make_sources(
                temp_path, temp_path / "Sources", ".", override_path
            )
            self.make_release(temp_path / "Release", suite_name)
            (temp_path / "Packages").unlink()
            (temp_path / "Sources").unlink()
            self.gpg_sign(
                private_key, temp_path / "Release", temp_path / "InRelease"
            )
            [dsc_name] = [
                p.name for p in temp_path.iterdir() if p.suffix == ".dsc"
            ]
            dsc_data = (temp_path / dsc_name).read_bytes()
            [deb_name] = [
                p.name for p in temp_path.iterdir() if p.suffix == ".deb"
            ]
            deb_data = (temp_path / deb_name).read_bytes()
            host, port = self.serve_archive(temp_path)

            result = DebusineServer.execute_command(
                "create_work_request",
                "--created-by",
                "test-user",
                "server",
                "aptmirror",
                stdin=yaml.safe_dump(
                    {
                        "collection": suite_name,
                        "url": f"http://{host}:{port}/",
                        "suite": "./",
                        "architectures": [self.architecture],
                        "signing_key": public_key.decode(),
                    }
                ),
            )
            self.assertEqual(result.returncode, 0)
            self.wait_and_assert_result(
                yaml.safe_load(result.stdout)["work_request_id"]
            )

        # We don't mirror indexes for flat repositories, but we can at least
        # fetch the packages directly (even though doing so via the pooled
        # layout doesn't really make sense).
        archive_url = f"https://deb.{socket.getfqdn()}/debusine/System"
        self.assertEqual(
            requests.get(
                f"{archive_url}/pool/main/d/debusine/{dsc_name}"
            ).content,
            dsc_data,
        )
        self.assertEqual(
            requests.get(
                f"{archive_url}/pool/main/d/debusine/{deb_name}"
            ).content,
            deb_data,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-admin command to manage workspaces.

Note: to make commands easier to be invoked from Ansible, we take care to make
them idempotent.
"""

import datetime as dt
from collections.abc import Callable
from typing import Any, NoReturn, cast, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser
from django.db import transaction

from debusine.artifacts.models import SINGLETON_COLLECTION_CATEGORIES
from debusine.db.context import context
from debusine.db.models import (
    Collection,
    Group,
    Scope,
    Workspace,
    default_workspace,
)
from debusine.db.models.workspaces import DeleteWorkspaces, WorkspaceRole
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import (
    ResourceRoles,
    Workspaces,
    get_scope,
    get_workspace,
)


class Command(DebusineBaseCommand):
    """Command to manage workspaces."""

    help = "Manage workspaces"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        define = subparsers.add_parser(
            "define",
            aliases=["create"],
            help="Create a workspace with defaults"
            " if it doesn't exist. Modify it with the specified"
            " parameters if it doesn't.",
        )
        define.add_argument(
            "scope_workspace",
            metavar="scope/name",
            help="scope/name for new workspace",
        )
        define.add_argument(
            "owners_group",
            nargs="?",
            help="Name of the owners group for the new workspace,"
            " mandatory when creating workspace",
        )
        public = define.add_mutually_exclusive_group()
        public.add_argument("--public", action="store_true", default=None)
        public.add_argument(
            "--private",
            action="store_false",
            dest="public",
            help='Public permissions (default: private)',
        )
        define.add_argument(
            "--default-expiration-delay",
            action="store",
            type=int,
            metavar="N",
            default=None,
            help="Minimal time (in days) that a new artifact is kept in the"
            " workspace before being expired, 0 means no expiration"
            " (default: 30)",
        )
        define.add_argument(
            "--no-singleton-collections",
            dest="singleton_collections",
            action="store_false",
            help=(
                "Don't create the usual singleton collections for this "
                "workspace (default: create singleton collections)"
            ),
        )

        rename = subparsers.add_parser(
            "rename", help=self.handle_rename.__doc__
        )
        rename.add_argument(
            "scope_workspace", metavar="scope/name", help="Workspace to rename"
        )
        rename.add_argument("new_name", help="New name for the workspace")

        delete = subparsers.add_parser(
            "delete", help=self.handle_delete.__doc__
        )
        delete.add_argument(
            "scope_workspace", metavar="scope/name", help="Workspace to delete"
        )
        delete.add_argument(
            '--yes', action='store_true', help='Skips confirmation of deletion'
        )
        delete.add_argument(
            "--force",
            action="store_true",
            help="Do not fail if trying to delete nonexistent workspaces",
        )

        list_sp = subparsers.add_parser("list", help=self.handle_list.__doc__)
        list_sp.add_argument(
            "--yaml", action="store_true", help="Machine readable YAML output"
        )
        list_sp.add_argument(
            "scope",
            help="list workspaces for this scope",
            nargs='?',
            default=None,
        )

        list_roles = subparsers.add_parser(
            "list_roles", help=self.handle_list_roles.__doc__
        )
        list_roles.add_argument(
            "scope_workspace", metavar="scope/name", help="Workspace to inspect"
        )
        list_roles.add_argument(
            "--yaml", action="store_true", help="Machine readable YAML output"
        )

        grant_role = subparsers.add_parser(
            "grant_role", help=self.handle_grant_role.__doc__
        )
        grant_role.add_argument(
            "scope_workspace", metavar="scope/name", help="Workspace to edit"
        )
        grant_role.add_argument("role", help="Name of the role to assign")
        grant_role.add_argument(
            "groups",
            nargs='+',
            metavar="group",
            help="Group(s) getting the role",
        )

        revoke_role = subparsers.add_parser(
            "revoke_role", help=self.handle_revoke_role.__doc__
        )
        revoke_role.add_argument(
            "scope_workspace", metavar="scope/name", help="Workspace to edit"
        )
        revoke_role.add_argument("role", help="Name of the role to revoke")
        revoke_role.add_argument(
            "groups",
            nargs='+',
            metavar="group",
            help="Group(s) dropping the role",
        )

    def get_scope_and_workspace_name(
        self, scope_workspace: str
    ) -> tuple[Scope, str]:
        """Lookup a scopename/workspacename string."""
        if "/" not in scope_workspace:
            raise CommandError(
                f"scope_workspace {scope_workspace!r} should be in the form"
                " 'scopename/workspacename'",
                returncode=3,
            )
        scope_name, workspace_name = scope_workspace.split("/", 1)
        return get_scope(scope_name), workspace_name

    def get_group(self, scope: Scope, group_name: str) -> Group:
        """Lookup a group in a scope."""
        try:
            group = Group.objects.get(scope=scope, name=group_name)
        except Group.DoesNotExist:
            raise CommandError(
                f"Group {group_name!r} not found in scope {scope.name!r}",
                returncode=3,
            )
        return group

    @override
    @context.disable_permission_checks()
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Dispatch the requested action."""
        func = cast(
            Callable[..., NoReturn],
            getattr(self, f"handle_{options['action']}", None),
        )
        func(*args, **options)

    def handle_define(
        self,
        *,
        scope_workspace: str,
        owners_group: str,
        **options: Any,
    ) -> NoReturn:
        """
        Ensure a workspace exists, with an Owners group.

        This is idempotent, and it makes sure the named workspace
        exists, it has an "Owners" group, and that group has the ADMIN
        role on the workspace.
        """
        scope, workspace_name = self.get_scope_and_workspace_name(
            scope_workspace
        )

        new_workspace = False
        with transaction.atomic():
            try:
                try:
                    workspace = Workspace.objects.get(
                        scope=scope, name=workspace_name
                    )
                except Workspace.DoesNotExist:
                    new_workspace = True
                    workspace = Workspace(
                        name=workspace_name,
                        scope=scope,
                        # defaults:
                        public=False,
                        default_expiration_delay=dt.timedelta(30),
                    )
                if options["public"] is not None:
                    workspace.public = options["public"]
                if options["default_expiration_delay"] is not None:
                    workspace.default_expiration_delay = dt.timedelta(
                        days=options["default_expiration_delay"]
                    )
                workspace.full_clean()
                workspace.save()
                if options["singleton_collections"]:
                    for category in SINGLETON_COLLECTION_CATEGORIES:
                        Collection.objects.get_or_create_singleton(
                            category, workspace
                        )

                # Make sure the workspace has an Owners group
                if new_workspace and owners_group is None:
                    raise CommandError(
                        "Error: owners_group is required"
                        " when creating workspace",
                        returncode=3,
                    )
                if owners_group is not None:
                    admin_group = self.get_group(scope, owners_group)
                    admin_group.assign_role(workspace, "owner")
            except ValidationError as exc:
                raise CommandError(
                    "Error creating workspace: " + "\n".join(exc.messages),
                    returncode=3,
                )

        raise SystemExit(0)

    handle_create = handle_define

    def handle_rename(
        self, *, scope_workspace: str, new_name: str, **options: Any
    ) -> NoReturn:
        """Rename a workspace."""
        workspace = get_workspace(scope_workspace)

        if new_name == workspace.name:
            raise SystemExit(0)

        workspace.name = new_name
        try:
            workspace.full_clean()
        except ValidationError as exc:
            self.stderr.write("Renamed workspace would be invalid:")
            for field, errors in exc.message_dict.items():
                for error in errors:
                    self.stderr.write(f"* {field}: {error}")
            raise SystemExit(3)

        workspace.save()

        raise SystemExit(0)

    def handle_list(self, **options: Any) -> NoReturn:
        """List workspaces in a scope."""
        workspaces = Workspace.objects.all()
        if options["scope"] is not None:
            scope = get_scope(options["scope"])
            workspaces = workspaces.filter(scope=scope)
        Workspaces(options["yaml"]).print(workspaces, self.stdout)
        raise SystemExit(0)

    def handle_list_roles(
        self, scope_workspace: str, **options: Any
    ) -> NoReturn:
        """List workspace's roles."""
        workspace = get_workspace(scope_workspace)

        workspace_roles = WorkspaceRole.objects.filter(resource=workspace)
        ResourceRoles(options["yaml"]).print(workspace_roles, self.stdout)
        raise SystemExit(0)

    def handle_grant_role(
        self, scope_workspace: str, role: str, groups: list[str], **options: Any
    ) -> NoReturn:
        """Assign workspace's role to groups."""
        workspace = get_workspace(scope_workspace)
        try:
            for group_name in groups:
                group = self.get_group(workspace.scope, group_name)
                group.assign_role(workspace, role)
        except ValueError as exc:
            raise CommandError(
                "Error assigning role: " + str(exc),
                returncode=3,
            )
        raise SystemExit(0)

    def handle_revoke_role(
        self, scope_workspace: str, role: str, groups: list[str], **options: Any
    ) -> NoReturn:
        """Revoke workspace's role from groups."""
        workspace = get_workspace(scope_workspace)
        for group_name in groups:
            group = self.get_group(workspace.scope, group_name)
            WorkspaceRole.objects.filter(
                resource=workspace, group=group, role=role
            ).delete()
        raise SystemExit(0)

    def handle_delete(
        self, *, scope_workspace: str, yes: bool, force: bool, **options: Any
    ) -> NoReturn:
        """Delete a workspace with associated resources."""
        try:
            workspace = get_workspace(scope_workspace)
        except CommandError:
            if force:
                raise SystemExit(0)
            else:
                raise

        # Prevent deletion of default workspace
        if workspace == default_workspace():
            raise CommandError(
                f"Workspace {scope_workspace} cannot be deleted",
                returncode=3,
            )

        deletion_confirmed = False
        if yes:
            deletion_confirmed = True
        else:
            deletion_answer = input(
                f"Would you like to delete workspace {scope_workspace}? [yN] "
            )
            deletion_confirmed = deletion_answer.strip() in ('y', 'Y')

        if not deletion_confirmed:
            raise SystemExit(0)

        with transaction.atomic():
            operation = DeleteWorkspaces(
                Workspace.objects.filter(pk=workspace.pk)
            )
            operation.perform_deletions()

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for base server-side task classes."""

import re
from pathlib import Path
from typing import ClassVar, override
from unittest import mock

import debusine.worker.tags as wtags
from debusine.artifacts.local_artifact import WorkRequestDebugLogs
from debusine.db.context import context
from debusine.db.models import Artifact, ArtifactRelation, WorkRequest
from debusine.db.models.permissions import format_permission_check_error
from debusine.db.permissioncontext import PermissionContext
from debusine.server.tasks.base import (
    ServerTaskPermissionDenied,
    get_provided_celery_worker_tags,
)
from debusine.server.tasks.tests.helpers import SampleBaseServerTask
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData, WorkerType
from debusine.test.django import TestCase
from debusine.worker.system_information import system_metadata


class SampleBaseServerTask1(
    SampleBaseServerTask[BaseTaskData, BaseDynamicTaskData]
):
    """Sample class to test BaseServerTask class."""

    @override
    def _execute(self) -> WorkRequest.Results:
        """Unused abstract method from DBTask."""
        raise NotImplementedError()


class SampleBaseServerTask2Data(BaseTaskData):
    """Data representation for SampleBaseServerTask2."""

    foo: str


class SampleBaseServerTask2(
    SampleBaseServerTask[SampleBaseServerTask2Data, BaseDynamicTaskData]
):
    """Test BaseServerTask class with jsonschema validation."""

    TASK_VERSION = 1

    @override
    def _execute(self) -> WorkRequest.Results:
        """Unused abstract method from DBTask."""
        raise NotImplementedError()


class BaseServerTaskTests(TestCase):
    """Unit tests for :py:class:`BaseServerTask`."""

    task: ClassVar[SampleBaseServerTask1]
    task2: ClassVar[SampleBaseServerTask2]
    worker_metadata: ClassVar[dict[str, WorkerType]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        task = cls.playground.create_server_task(
            "samplebaseservertask1", {}
        ).get_task()
        assert isinstance(task, SampleBaseServerTask1)
        cls.task = task
        task2 = cls.playground.create_server_task(
            "samplebaseservertask2", {"foo": "bar"}
        ).get_task()
        assert isinstance(task2, SampleBaseServerTask2)
        cls.task2 = task2
        cls.worker_metadata = system_metadata(WorkerType.CELERY)

    def test_can_run_on_no_version(self) -> None:
        """Ensure can_run_on returns True if no version is specified."""
        self.assertIsNone(self.task.TASK_VERSION)
        metadata = {**self.worker_metadata, **self.task.analyze_worker()}
        self.assertEqual(self.task.can_run_on(metadata), True)

    def test_can_run_on_with_different_versions(self) -> None:
        """Ensure can_run_on returns False if versions differ."""
        self.assertIsNone(self.task.TASK_VERSION)
        metadata = {**self.worker_metadata, **self.task.analyze_worker()}
        metadata["server:samplebaseservertask1:version"] = 1
        self.assertEqual(self.task.can_run_on(metadata), False)

    def test_get_provided_celery_worker_tags(self) -> None:
        tags = get_provided_celery_worker_tags()
        self.assertIn(
            "worker:task:server:servernoop:version:1",
            tags,
        )

        # Ensure only server tasks are included
        for worker_type in (
            "worker",
            "internal",
            "workflow",
            "signing",
            "wait",
        ):
            with self.subTest(worker_type=worker_type):
                self.assertEqual(
                    [
                        tag
                        for tag in tags
                        if tag.startswith(wtags.TASK_PREFIX + worker_type + ":")
                    ],
                    [],
                )

    def test_execute_sets_up_context(self) -> None:
        """`execute` sets up a suitable context."""

        def fake_execute() -> WorkRequest.Results:
            self.assertEqual(
                context.scope, self.task.work_request.workspace.scope
            )
            self.assertEqual(context.user, self.task.work_request.created_by)
            self.assertEqual(
                context.workspace, self.task.work_request.workspace
            )
            return WorkRequest.Results.SUCCESS

        with mock.patch.object(
            SampleBaseServerTask1, "_execute", side_effect=fake_execute
        ):
            self.task.execute()

        self.assertIsNone(context.scope)
        self.assertIsNone(context._permission_context.get())
        self.assertIsNone(context.workspace)

    def setup_upload_work_request_debug_logs(
        self, source_artifacts: list[Artifact] | None = None
    ) -> list[tuple[str, bytes]]:
        """Setup for upload_work_request_debug_logs tests."""  # noqa: D401
        # Add a file to be uploaded
        with self.task.open_debug_log_file("test.log") as file:
            file.write("log")
        assert self.task._debug_log_files_directory is not None

        self.task._source_artifacts_ids = [
            artifact.id for artifact in source_artifacts or []
        ]

        return [
            (path.name, path.read_bytes())
            for path in Path(
                self.task._debug_log_files_directory.name
            ).iterdir()
        ]

    def assert_uploaded_work_request_debug_logs_artifact(
        self,
        work_request: WorkRequest,
        expected_files: list[tuple[str, bytes]],
    ) -> Artifact:
        """Assert that an Artifact was created."""
        artifact = Artifact.objects.filter(
            category=WorkRequestDebugLogs._category
        ).last()
        assert artifact is not None
        assert artifact.created_by_work_request is not None

        self.assertEqual(artifact.category, WorkRequestDebugLogs._category)
        self.assertEqual(artifact.workspace, work_request.workspace)
        self.assertEqual(artifact.data, {})
        self.assertEqual(artifact.created_by_work_request, work_request)
        for file_in_artifact, (name, contents) in zip(
            artifact.fileinartifact_set.order_by("id"), expected_files
        ):
            self.assertEqual(file_in_artifact.path, name)
            file_backend = work_request.workspace.scope.download_file_backend(
                file_in_artifact.file
            )
            with file_backend.get_stream(file_in_artifact.file) as file:
                self.assertEqual(file.read(), contents)

        return artifact

    @context.disable_permission_checks()
    def test_upload_work_request_debug_logs_with_relation(self) -> None:
        """
        Artifact is created and uploaded. Relation is created.

        The relation is from the debug logs artifact to the source_artifact_id.
        """
        source_artifact = self.playground.create_artifact()[0]
        expected_files = self.setup_upload_work_request_debug_logs(
            [source_artifact]
        )
        assert self.task.work_request is not None

        self.task._upload_work_request_debug_logs()

        artifact = self.assert_uploaded_work_request_debug_logs_artifact(
            self.task.work_request, expected_files
        )

        self.assertEqual(artifact.relations.count(), 1)
        relation = artifact.relations.first()
        assert relation is not None
        self.assertEqual(relation.artifact, artifact)
        self.assertEqual(relation.target, source_artifact)
        self.assertEqual(relation.type, ArtifactRelation.Relations.RELATES_TO)

    @context.disable_permission_checks()
    def test_upload_work_request_debug_logs(self) -> None:
        """Artifact is created and uploaded."""
        expected_files = self.setup_upload_work_request_debug_logs()
        assert self.task.work_request is not None

        self.task._upload_work_request_debug_logs()

        self.assert_uploaded_work_request_debug_logs_artifact(
            self.task.work_request, expected_files
        )

    def test_upload_work_request_no_log_files(self) -> None:
        """No log files: no artifact created."""
        self.task._upload_work_request_debug_logs()

        self.assertFalse(
            Artifact.objects.filter(
                category=WorkRequestDebugLogs._category
            ).exists()
        )

    def test_enforce(self) -> None:
        """Test enforcing permissions."""
        self.task.work_request.set_current()

        workspace = self.playground.get_default_workspace()

        self.task.enforce(workspace.can_display)

        with self.assertRaisesRegex(
            ServerTaskPermissionDenied,
            re.escape(
                format_permission_check_error(
                    workspace.can_create_work_requests,
                    PermissionContext.for_work_request(
                        work_request=self.task.work_request
                    ),
                )
            ),
        ):
            self.task.enforce(workspace.can_create_work_requests)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine template context processors."""

from typing import Any, TYPE_CHECKING

from django.conf import settings

if TYPE_CHECKING:  # pragma: no cover
    import django.http


def server_info(
    request: "django.http.HttpRequest",  # noqa: ARG001, U100
) -> dict[str, Any]:
    """Add general server information to the request context."""
    return {
        "DEBUSINE_FQDN": settings.DEBUSINE_FQDN,
    }


def application_context(
    request: "django.http.HttpRequest",  # noqa: ARG001, U100
) -> dict[str, Any]:
    """Add general server information to the request context."""
    from debusine.db.context import context

    return {
        "scope": lambda: context.scope,
        "workspace": lambda: context.workspace,
    }



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Populate ``worker:build-arch:`` tags from work request data."""

from typing import Any

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps

BUILD_ARCH_PREFIX = "worker:build-arch:"

GENERATED_TAGS = [
    BUILD_ARCH_PREFIX + arch
    for arch in ("amd64", "arm64", "ppc64el", "armhf", "armel", "i386")
]


def copy_build_arch_to_tag(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Convert work request build architectures to tags."""
    WorkRequest = apps.get_model("db", "WorkRequest")

    def merge_tag_into(tag: str, wr: Any) -> None:
        """Add tag to tags, in place, keeping it sorted."""
        assert isinstance(wr, WorkRequest)
        wr.scheduler_tags_required.append(tag)
        wr.scheduler_tags_required.sort()
        wr.save()

    # sbuild-like tags using build_architecture
    for wr in (
        WorkRequest.objects.filter(task_data__has_key="build_architecture")
        .exclude(task_data__build_architecture=None)
        .exclude(scheduler_tags_required__overlap=GENERATED_TAGS)
    ):
        tag = BUILD_ARCH_PREFIX + wr.task_data["build_architecture"]
        merge_tag_into(tag, wr)

    # mmdebstrap-like tags using bootstrap_options.architecture
    for wr in (
        WorkRequest.objects.filter(
            task_data__bootstrap_options__has_key="architecture"
        )
        .exclude(task_data__bootstrap_options__architecture=None)
        .exclude(scheduler_tags_required__overlap=GENERATED_TAGS)
    ):
        tag = (
            BUILD_ARCH_PREFIX
            + wr.task_data["bootstrap_options"]["architecture"]
        )
        merge_tag_into(tag, wr)


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0045_alter_workrequest_unblock_strategy'),
    ]

    operations = [
        migrations.RunPython(copy_build_arch_to_tag, migrations.RunPython.noop),
    ]
